package de.mpg.mpimet.yacgui.utils;

import java.util.ArrayList;
import java.util.List;

import de.mpg.mpimet.yacgui.xml.Calendar;
import de.mpg.mpimet.yacgui.xml.CouplingOperation;
import de.mpg.mpimet.yacgui.xml.Dates;
import de.mpg.mpimet.yacgui.xml.Dom;
import de.mpg.mpimet.yacgui.xml.TimestepUnit;
import de.mpg.mpimet.yacgui.xml.interpolation.Interpolation;
import de.mpg.mpimet.yacgui.xml.interpolation.InterpolationFactory;
import de.mpg.mpimet.yacgui.xml.interpolation.InterpolationMethod;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

public class Default {

	public static final String DATE_FORMAT = "Gyyyy-MM-dd'T'HH:mm:ss.SSS";
	public static final String CREATED_DATE_FORMAT = "dd-MM-yyyy HH:mm";
	
	/**
	 * Start date for coupling, should be written according {@link Default#DATE_FORMAT}
	 */
	public static final String START_DATE = "1800-01-01T00:00:00.000";
	/**
	 * End date for coupling, should be written according {@link Default#DATE_FORMAT}
	 */
	public static final String END_DATE = "2100-01-01T00:00:00.000";
	
	public static final Calendar CALENDAR = Calendar.PROLEPTIC_GREGORIAN;
	public static final Dates DATES = new Dates(START_DATE, END_DATE, CALENDAR);

	public static final TimestepUnit TIMESTEP_UNIT = TimestepUnit.SECOND;

	public static final String TIMESTEP_SOURCE = "10";
	public static final String TIMESTEP_TARGET = "10";
	public static final String COUPLING_PERIOD = "60";
	public static final CouplingOperation COUPLING_OPERATION = CouplingOperation.ACCUMULATE;
	public static final String SOURCE_TIMELAG = "0";
	public static final String TARGET_TIMELAG = "0";

	public static final Double MASK_VALUE = null;

	public static final boolean AT_SOURCE_BEFORE_INTERPOLATION = false;
	public static final boolean AT_SOURCE_AFTER_INTERPOLATION = false;
	public static final boolean AT_TARGET = false;

	public static final String INTERPOLATION_ATTR_N = "1";
	public static final String INTERPOLATION_ATTR_WEIGHTED = Dom.Value.ARITHMETIC_AVERAGE;
	public static final String INTERPOLATION_ATTR_RBF_N = "9";
	public static final String INTERPOLATION_ATTR_RBF_SCALE = "14.87973";
	public static final String INTERPOLATION_ATTR_GAUSS_SCALE = "0.1";
	public static final String INTERPOLATION_ATTR_RBF_KERNEL = Dom.Value.GAUSS_KERNEL;
	
	public static final String INTERPOLATION_ATTR_CONSERVATIVE_ORDER = "1";
	public static final String INTERPOLATION_ATTR_ENFORCED_CONSERVATION = "false";
	public static final String INTERPOLATION_ATTR_PARTIAL_COVERAGE = "false";
	public static final String INTERPOLATION_ATTR_NORMALISATION = Dom.Value.DESTAREA;
	
	public static final String INTERPOLATION_ATTR_CORNER_WEIGHT = "1";

	public static final String INTERPOLATION_ATTR_USER_VALUE = "0.0";
	public static final String INTERPOLATION_ATTR_SPREAD_DISTANCE = "0.0";
	public static final String INTERPOLATION_ATTR_MAX_SEARCH_DISTANCE = "0.0";

	public static final String INTERPOLATION_ATTR_CREEP_DISTANCE = "-1";

	/**
	 * Creates an implementation of {@link Interpolation} according to 
	 * given method, and uses default values as parameters. Sequence of
	 * attributes must be according to alphabetical order of attribute
	 * string values! see class {@link Dom}  
	 * 
	 * @param method	String representation of {@link InterpolationMethod}
	 * @return Interpolation with default values as parameters.
	 */
	public static Interpolation defaultInterpolation(String method) {
		if (!InterpolationMethod.isValidName(method)) {
			throw new IllegalArgumentException("'" + method + "' is not a valid method");
		}
		InterpolationMethod interpolationMethod = InterpolationMethod.fromName(method);

		List<String[]> attributes = null;

		switch (interpolationMethod) {
		case AVERAGE:
			attributes = new ArrayList<String[]>();
			attributes.add(new String[] {Dom.Attribute.PARTIAL_COVERAGE, INTERPOLATION_ATTR_PARTIAL_COVERAGE});
			attributes.add(new String[] {Dom.Attribute.WEIGHTED, INTERPOLATION_ATTR_WEIGHTED});
			break;
		case N_NEAREST_NEIGHBOR:
			attributes = new ArrayList<String[]>();
			attributes.add(new String[] {Dom.Attribute.GAUSS_SCALE, INTERPOLATION_ATTR_GAUSS_SCALE});
			attributes.add(new String[] {Dom.Attribute.N, INTERPOLATION_ATTR_N});
			attributes.add(new String[] {Dom.Attribute.WEIGHTED, INTERPOLATION_ATTR_WEIGHTED});
			break;
		case RADIAL_BASIS_FUNCTION:
			attributes = new ArrayList<String[]>();
			attributes.add(new String[] {Dom.Attribute.N, INTERPOLATION_ATTR_RBF_N});
			attributes.add(new String[] {Dom.Attribute.RBF_KERNEL, INTERPOLATION_ATTR_RBF_KERNEL});
			attributes.add(new String[] {Dom.Attribute.RBF_SCALE, INTERPOLATION_ATTR_RBF_SCALE});
			break;
		case CONSERVATIVE:
			attributes = new ArrayList<String[]>();
			attributes.add(new String[] {Dom.Attribute.ENFORCED_CONSERVATION, INTERPOLATION_ATTR_ENFORCED_CONSERVATION});
			attributes.add(new String[] {Dom.Attribute.NORMALISATION, INTERPOLATION_ATTR_NORMALISATION});
			attributes.add(new String[] {Dom.Attribute.CONSERVATIVE_ORDER, INTERPOLATION_ATTR_CONSERVATIVE_ORDER});
			attributes.add(new String[] {Dom.Attribute.PARTIAL_COVERAGE, INTERPOLATION_ATTR_PARTIAL_COVERAGE});
			break;
        	case BERNSTEIN_BEZIER:
            		attributes = new ArrayList<String[]>();
			break;
		case CREEP:
			attributes = new ArrayList<String[]>();
			attributes.add(new String[] {Dom.Attribute.CREEP_DISTANCE, INTERPOLATION_ATTR_CREEP_DISTANCE});
			break;
		case FIXED_VALUE:
			attributes = new ArrayList<String[]>();
			attributes.add(new String[] {Dom.Attribute.USER_VALUE, INTERPOLATION_ATTR_USER_VALUE});
			break;
		case USER_FILE:
			attributes = new ArrayList<String[]>();
			attributes.add(new String[] {Dom.Attribute.FILENAME, null });
			break;
		case SOURCE2TARGET_MAP:
			attributes = new ArrayList<String[]>();
			attributes.add(new String[] {Dom.Attribute.SPREAD_DISTANCE, INTERPOLATION_ATTR_SPREAD_DISTANCE});
			attributes.add(new String[] {Dom.Attribute.MAX_SEARCH_DISTANCE, INTERPOLATION_ATTR_MAX_SEARCH_DISTANCE});
			attributes.add(new String[] {Dom.Attribute.WEIGHTED, INTERPOLATION_ATTR_WEIGHTED});
			break;
		default:
			throw new IllegalArgumentException("'" + method + "' is not a valid method");
		}

		return InterpolationFactory.createInterpolation(method, attributes);
	}

}
