package de.mpg.mpimet.yacgui.utils;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import de.mpg.mpimet.yacgui.xml.Calendar;
import de.mpg.mpimet.yacgui.xml.Component;
import de.mpg.mpimet.yacgui.xml.Couple;
import de.mpg.mpimet.yacgui.xml.Coupling;
import de.mpg.mpimet.yacgui.xml.Dates;
import de.mpg.mpimet.yacgui.xml.Dom;
import de.mpg.mpimet.yacgui.xml.Dom.Attribute;
import de.mpg.mpimet.yacgui.xml.Grid;
import de.mpg.mpimet.yacgui.xml.Timestep;
import de.mpg.mpimet.yacgui.xml.TimestepUnit;
import de.mpg.mpimet.yacgui.xml.Transient;
import de.mpg.mpimet.yacgui.xml.TransientCouple;
import de.mpg.mpimet.yacgui.xml.TransientGridRef;
import de.mpg.mpimet.yacgui.xml.interpolation.Interpolation;
import de.mpg.mpimet.yacgui.xml.interpolation.InterpolationFactory;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Main store/parse class. Responsible for parsing coupling, components, as well storing coupling.
 * 
 * @author Maxim Yastremsky
 */
public class XmlDomHandler {
	
	private static final String JAXP_SCHEMA_LANGUAGE = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
	private static final String W3C_XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";
	
	public enum XmlType {
		COUPLING(Dom.COUPLING),
		COMPONENT(Dom.COMPONENT);
		
		private String rootElement;
		
		private XmlType(String rootElement) {
			this.rootElement = rootElement;
		}
		
		public String getRootElement() {
			return rootElement;
		}
	}

	protected Document document;
	private XmlDomErrorHandler xmlDomErrorHandler;
	private boolean isXsdFileAvailable;

	private void checkRootTagName(XmlType type) throws XmlDomHandlerException {
		String rootName = document.getDocumentElement().getNodeName();
		
		if (!type.getRootElement().equals(rootName)) {
			String message = "Root element of xml file should be '" + type.getRootElement() + "', but it is '" + rootName + "'.";
			throw new XmlDomHandlerException(message);
		}
	}
	
	/**
	 * @param filename	path to xml file to be parsed
	 * @param type      componenent of coupling xml file
	 * @throws XmlDomHandlerException if file is malformed (does not validate)
	 */
	public XmlDomHandler(String filename, XmlType type) throws XmlDomHandlerException {
		parseXml(filename, type);
	}
	
	/**
	 * Parses provided xml file and validates it according founded xsd file.
	 * 
	 * @param filename path to xml file to be parsed
	 * @param type     componenent of coupling xml file
	 * @throws  XmlDomHandlerException if file is malformed (does not validate)
	 */
	public void parseXml(String filename, XmlType type) throws XmlDomHandlerException {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		
		if (type != null) {
			factory.setValidating(true);
			factory.setNamespaceAware(true);
			factory.setAttribute(JAXP_SCHEMA_LANGUAGE, W3C_XML_SCHEMA);
		}
			
		try {
			DocumentBuilder builder = factory.newDocumentBuilder();
			xmlDomErrorHandler = new XmlDomErrorHandler();
			builder.setErrorHandler(xmlDomErrorHandler);
			document = builder.parse(filename);

			checkRootTagName(type);
			checkCorrespondingXsdFile(filename);
			
		} catch (ParserConfigurationException e) {
			Logger.logError("ParserConfigurationException during DocumentBuilder.parse(..)", e);
			throw new XmlDomHandlerException(e.getMessage());
		} catch (SAXException e) {
			Logger.logError("SAXException during DocumentBuilder.parse(..)", e);
			throw new XmlDomHandlerException(e.getMessage());
		} catch (IOException e) {
			Logger.logError("IOException during DocumentBuilder.parse(..)", e);
			throw new XmlDomHandlerException(e.getMessage());
		}
	}
	
	private void checkCorrespondingXsdFile(String filename) {
		//get the root element and extract from it the xsd file name
		Element element = document.getDocumentElement();
		String attribute = element.getAttribute(Dom.Attribute.XSI_SCHEMALOCATION);
		String[] parts = null;
		
		//this implementation is not really clean
		//should have the value: "http://www.w3schools.com <file-name>.xsd"
		if (attribute != null && (parts = attribute.split(" ")).length == 2) {
			String xsdFilePath = new File(filename).getParent() + "/" + parts[1];
			isXsdFileAvailable = new File(xsdFilePath).exists();
		}
	}

	/**
	 * Returns error messages from xsd validation, in case if corresponding xsd file was found,
	 * otherwise null;
	 * 
	 * @return xsd validation messages
	 */
	public String getErrorHandlerMessages() {
		if (xmlDomErrorHandler == null || !isXsdFileAvailable) {
			return null;
		} else {
			return xmlDomErrorHandler.getStoredErrorMessages();
		}
	}

	/**
	 * Builds a component from previously parsed xml file.
	 * 
	 * @return a component
	 */
	public Component buildComponent() {
		
		Element componentElement = (Element) document.getElementsByTagName(Dom.COMPONENT).item(0);
		Component component = parseComponentCommonInfo(componentElement);
		
		// <transients>
		Map<Integer, Transient> transients = parseTransients(document.getElementsByTagName(Dom.TRANSIENT));
		component.setTransients(transients);

		// <grids>
		Map<Integer, Grid> grids = parseGrids(document.getElementsByTagName(Dom.GRID));
		component.setGrids(grids);
		return component;
	}

	private Component parseComponentCommonInfo(Element componentElement) {
		//general component info
		int id = -1;
		if (componentElement.hasAttribute(Dom.Attribute.ID)) { // couple case
			id = Integer.parseInt(componentElement.getAttribute(Dom.Attribute.ID));
		} else {		//single component case
			id = Integer.parseInt(componentElement.getElementsByTagName(Dom.ID).item(0).getTextContent());
		}
		String name = componentElement.getElementsByTagName(Dom.NAME).item(0).getTextContent();
		String model = componentElement.getElementsByTagName(Dom.MODEL).item(0).getTextContent();
		String simulated = componentElement.getElementsByTagName(Dom.SIMULATED).item(0).getTextContent();

		// <transient_grid_refs>
		NodeList transientNodes = componentElement.getElementsByTagName(Dom.TRANSIENT_GRID_REF);
		Map<Integer, TransientGridRef> transientGridRefs = new HashMap<Integer, TransientGridRef>();
		for (int j = 0; j < transientNodes.getLength(); j++) {
			Element transientRefElement = (Element) transientNodes.item(j);
			
			int tr_id = Integer.parseInt(transientRefElement.getAttribute(Dom.Attribute.ID));
			int transientRef = Integer.parseInt(transientRefElement.getAttribute(Dom.Attribute.TRANSIENT_REF));
			int gridRef = Integer.parseInt(transientRefElement.getAttribute(Dom.Attribute.GRID_REF));
			int collectionSize = Integer.parseInt(transientRefElement.getAttribute(Dom.Attribute.COLLECTION_SIZE));
			TransientGridRef transientGridRef = new TransientGridRef(tr_id, transientRef, gridRef, collectionSize);
			transientGridRefs.put(tr_id, transientGridRef);
		}

		return new Component(id, name, model, simulated, transientGridRefs);
	}

	private Map<Integer, Grid> parseGrids(NodeList gridsNodeList) {
		Map<Integer, Grid> grids = new HashMap<Integer, Grid>();
		for (int i = 0; i < gridsNodeList.getLength(); i++) {
			Element gridElement = (Element) gridsNodeList.item(i);
			int gridId = Integer.parseInt(gridElement.getAttribute(Dom.Attribute.ID));
			String aliasName = gridElement.getAttribute(Dom.Attribute.ALIAS_NAME);
			Grid grid = new Grid(gridId, aliasName);
			grids.put(gridId, grid);
		}
		return grids;
	}

	private Map<Integer, Transient> parseTransients(NodeList transientsNodeList) {
		Map<Integer, Transient> transients = new HashMap<Integer, Transient>();
		for (int i = 0; i < transientsNodeList.getLength(); i++) {
			Element transientElement = (Element) transientsNodeList.item(i);
			int transientId = Integer.parseInt(transientElement.getAttribute(Dom.Attribute.ID));
			String standardNameAttribute = transientElement.getAttribute(Dom.Attribute.TRANSIENT_STANDARD_NAME);
			Transient newTransient = new Transient(transientId, standardNameAttribute);
			transients.put(transientId, newTransient);
		}
		return transients;
	}

	/**
	 * Builds a Coupling from previously parsed xml file.
	 * 
	 * @return built Coupling
	 */
	public Coupling buildCoupling() {
		// <redirect properties>
		Element redirectElement = (Element) document.getElementsByTagName(Dom.REDIRECT).item(0);
		boolean stdoutRedirect = false;
		boolean rootRedirect = false;
		
		if (redirectElement != null) {
			stdoutRedirect = Boolean.parseBoolean(redirectElement.getAttribute(Attribute.REDIRECT_STDOUT));
			rootRedirect = Boolean.parseBoolean(redirectElement.getAttribute(Attribute.REDIRECT_OF_ROOT));
		}
		
		// <components>
		NodeList componentNodes = document.getElementsByTagName(Dom.COMPONENT);
		Map<Integer, Component> components = new HashMap<Integer, Component>();
		for (int i = 0; i < componentNodes.getLength(); i++) {
			Element componentElement = (Element) componentNodes.item(i);
			
			Component component = parseComponentCommonInfo(componentElement);
			components.put(component.getId(), component);
		}

		// <transients>
		Map<Integer, Transient> transients = parseTransients(document.getElementsByTagName(Dom.TRANSIENT)); 

		// <grids>
		Map<Integer, Grid> grids = parseGrids(document.getElementsByTagName(Dom.GRID));
		
		// <dates>
		Element datesElement = (Element) document.getElementsByTagName(Dom.DATES).item(0);
		Element startDateElement = (Element) datesElement.getElementsByTagName(Dom.START_DATE).item(0);
		String startDate = startDateElement.getTextContent();
		Element endDateElement = (Element) datesElement.getElementsByTagName(Dom.END_DATE).item(0);
		String endDate = endDateElement.getTextContent();
		Element calendarElement = (Element) datesElement.getElementsByTagName(Dom.CALENDAR).item(0);
		Calendar calendar = Calendar.objectFromString(calendarElement.getTextContent());
		Dates dates = new Dates(startDate, endDate, calendar);

		// <timestep_unit>
		Element timestepUnitElement = (Element) document.getElementsByTagName(Dom.TIMESTEP_UNIT).item(0);
		String timestepUnitValue = timestepUnitElement.getTextContent();
		TimestepUnit timestepUnit = TimestepUnit.fromString(timestepUnitValue);

		// <couples>
		Element couplesElement = (Element) document.getElementsByTagName(Dom.COUPLES).item(0);
		NodeList coupleNodes = couplesElement.getElementsByTagName(Dom.COUPLE);
		
		Map<String, Couple> couples = new HashMap<String, Couple>();
		for (int i = 0; i < coupleNodes.getLength(); i++) {
			Element coupleElement = (Element) coupleNodes.item(i);
			Couple couple = parseCouple(coupleElement);
			couples.put(Utils.getCoupleId(couple.getComponent1Id(), couple.getComponent2Id()), couple);
		}
		
		Coupling coupling = new Coupling(stdoutRedirect, rootRedirect, components, transients, grids, dates, timestepUnit, couples);
		return coupling;
	}

	private Couple parseCouple(Element coupleElement) {
		// <component1>
		Element component1Element = (Element) coupleElement.getElementsByTagName(Dom.COMPONENT1).item(0);
		int component1Id = Integer.parseInt(component1Element.getAttribute(Dom.Attribute.COMPONENT_ID));

		// <component2>
		Element component2Element = (Element) coupleElement.getElementsByTagName(Dom.COMPONENT2).item(0);
		int component2Id = Integer.parseInt(component2Element.getAttribute(Dom.Attribute.COMPONENT_ID));

		// <transient_couple>
		List<TransientCouple> transientCouples = new ArrayList<TransientCouple>();
		NodeList transientCoupleNodes = coupleElement.getElementsByTagName(Dom.TRANSIENT_COUPLE);
		for (int j = 0; j < transientCoupleNodes.getLength(); j++) {
			Element transientCoupleElement = (Element) transientCoupleNodes.item(j);
			int transientId = Integer.parseInt(transientCoupleElement.getAttribute(Dom.Attribute.TRANSIENT_ID));

			// <source>
			Element sourceElement = (Element) transientCoupleElement.getElementsByTagName(Dom.SOURCE).item(0);
			int sourceComponentRef = Integer.parseInt(sourceElement.getAttribute(Dom.Attribute.COMPONENT_REF));
			int sourceTransientGridRef = Integer.parseInt(sourceElement.getAttribute(Dom.Attribute.TRANSIENT_GRID_REF));

			// <target>
			Element targetElement = (Element) transientCoupleElement.getElementsByTagName(Dom.TARGET).item(0);
			int targetTransientGridRef = Integer.parseInt(targetElement.getAttribute(Dom.Attribute.TRANSIENT_GRID_REF));

			// <timestep>
			Element timestepElement = (Element) transientCoupleElement.getElementsByTagName(Dom.TIMESTEP).item(0);
			
			Element sourceTimestepElement = (Element) timestepElement.getElementsByTagName(Dom.SOURCE).item(0);
			String sourceTimestep = sourceTimestepElement.getTextContent();
			
			Element targetTimestepElement = (Element) timestepElement.getElementsByTagName(Dom.TARGET).item(0);
			String targetTimestep = targetTimestepElement.getTextContent();
			
			Element couplingPeriodElement = (Element) timestepElement.getElementsByTagName(Dom.COUPLING_PERIOD).item(0);
			String couplingPeriod = couplingPeriodElement.getTextContent();

			String couplingPeriodOperation = couplingPeriodElement.getAttribute(Dom.Attribute.OPERATION);
			
			Element sourceTimeLagElement = (Element) timestepElement.getElementsByTagName(Dom.SOURCE_TIMELAG).item(0);
			String sourceTimeLag = sourceTimeLagElement.getTextContent();
			//int sourceTimeLag = Integer.parseInt(sourceTimeLagElement.getTextContent());

			Element targetTimeLagElement = (Element) timestepElement.getElementsByTagName(Dom.TARGET_TIMELAG).item(0);
			String targetTimeLag = targetTimeLagElement.getTextContent();
			//int targetTimeLag = Integer.parseInt(targetTimeLagElement.getTextContent());
			
			Timestep timestep = new Timestep(sourceTimestep, targetTimestep, couplingPeriod,
					couplingPeriodOperation, sourceTimeLag, targetTimeLag);

			// <interpolation_requirements>
			Element interpolationRequirementsElement = (Element) transientCoupleElement
					.getElementsByTagName(Dom.INTERPOLATION_REQUIREMENTS).item(0);

			// <interpolation>
			List<Interpolation> interpolations = new ArrayList<Interpolation>();
			NodeList interpolationNodes = interpolationRequirementsElement.getElementsByTagName(Dom.INTERPOLATION);
			for (int k = 0; k < interpolationNodes.getLength(); k++) {
				Element interpolationElement = (Element) interpolationNodes.item(k);
				String method = interpolationElement.getAttribute(Dom.Attribute.METHOD);

				List<String[]> attributes = new ArrayList<String[]>();
				NamedNodeMap attributeNodes = interpolationElement.getAttributes();
				
				int attrNumber = attributeNodes.getLength(); // number = method + attributes
				
				if (attrNumber > 1) {
					
					for (int l = 0; l < attrNumber; l++) { // cycle for all attributes
						Node attributeNode = attributeNodes.item(l);
						String attributeName = attributeNode.getNodeName();
						// skip "method"
						if (attributeName.equals(Dom.Attribute.METHOD)) {
							continue;
						}
						
						String attributeValue = attributeNode.getTextContent();
						String[] attribute = new String[] {attributeName, attributeValue};
						attributes.add(attribute);
					}
				}

				Interpolation interpolation = InterpolationFactory.createInterpolation(method, attributes);
				interpolations.add(interpolation);
			}

			// <mapping_on_source>
			boolean mappingOnSource = true;
			Element mappingOnSourceElement = (Element) transientCoupleElement
					.getElementsByTagName(Dom.MAPPING_ON_SOURCE).item(0);
			if (mappingOnSourceElement != null) {
				mappingOnSource = Boolean.parseBoolean(mappingOnSourceElement.getTextContent());
			}

			//<enforce_write_weight_file>
			Element enforceWriteWeightFileElement = (Element) transientCoupleElement
					.getElementsByTagName(Dom.ENFORCE_WRITE_WEIGHT_FILE).item(0);
			boolean enforceWriteWeightFile = Boolean.parseBoolean(enforceWriteWeightFileElement.getTextContent());
			String weightFilename = enforceWriteWeightFileElement.getAttribute(Dom.Attribute.FILENAME);

			TransientCouple transientCouple = new TransientCouple(transientId, sourceComponentRef, sourceTransientGridRef, 
					targetTransientGridRef, timestep, mappingOnSource, interpolations,
					enforceWriteWeightFile, weightFilename);
			transientCouples.add(transientCouple);
		}

		return new Couple(component1Id, component2Id, transientCouples);
	}

	/**
	 * Writes given {@link Coupling} to given File.
	 * 
	 * @param destination	the File to write to, must not be null
	 * @param coupling		the Coupling to write to File
	 */
	public static void writeCouplingToFile(File destination, Coupling coupling) {
		
		if (destination == null) {
			throw new IllegalArgumentException("destination file = null");
		}

		try {
			// instances for building the xml document
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();

			// building the tree, root element: <coupling>
			Document doc = docBuilder.newDocument();
			Element rootElement = doc.createElement(Dom.COUPLING);
			doc.appendChild(rootElement);
			rootElement.setAttribute(Dom.Attribute.XMLNS, Dom.Value.XMLNS);
			rootElement.setAttribute(Dom.Attribute.XMLNS_XSI, Dom.Value.XMLNS_XSI);
			rootElement.setAttribute(Dom.Attribute.XSI_SCHEMALOCATION, Dom.Value.XSI_SCHEMALOCATION);
			
			// first level elements:
			// <redirect>, <components>, <transients>, <grids>, <dates>, <timestep_unit>,
			// <couples>
			Element redirectElement = doc.createElement(Dom.REDIRECT);
			rootElement.appendChild(redirectElement);
			Element componentsElement = doc.createElement(Dom.COMPONENTS);
			rootElement.appendChild(componentsElement);
			Element transientsElement = doc.createElement(Dom.TRANSIENTS);
			rootElement.appendChild(transientsElement);
			Element gridsElement = doc.createElement(Dom.GRIDS);
			rootElement.appendChild(gridsElement);
			Element datesElement = doc.createElement(Dom.DATES);
			rootElement.appendChild(datesElement);
			Element timestepUnitElement = doc.createElement(Dom.TIMESTEP_UNIT);
			rootElement.appendChild(timestepUnitElement);
			Element couplesElement = doc.createElement(Dom.COUPLES);
			rootElement.appendChild(couplesElement);

			//redirect
			redirectElement.setAttribute(Dom.Attribute.REDIRECT_STDOUT, String.valueOf(coupling.isStdoutRedirect()));
			redirectElement.setAttribute(Dom.Attribute.REDIRECT_OF_ROOT, String.valueOf(coupling.isRootRedirect()));
			
			// <component>
			List<Component> components = new ArrayList<Component>(coupling.getComponentsMap().values());
			Utils.sortByComponentId(components);
			for (Component component : components) {
				Element componentElement = doc.createElement(Dom.COMPONENT);
				componentsElement.appendChild(componentElement);
				
				// <id>
				int id = component.getId();
				componentElement.setAttribute(Dom.Attribute.ID, String.valueOf(id));
				
				// <name>
				String name = component.getName();
				Element nameElement = doc.createElement(Dom.NAME);
				nameElement.setTextContent(name);
				componentElement.appendChild(nameElement);
				
				// <model>
				String model = component.getModel();
				Element modelElement = doc.createElement(Dom.MODEL);
				modelElement.setTextContent(model);
				componentElement.appendChild(modelElement);
				
				// <simulated>
				String simulated = component.getSimulated();
				Element simulatedElement = doc.createElement(Dom.SIMULATED);
				simulatedElement.setTextContent(simulated);
				componentElement.appendChild(simulatedElement);
				
				// <transient_grid_refs>
				Element transientGridRefsElement = doc.createElement(Dom.TRANSIENT_GRID_REFS);
				componentElement.appendChild(transientGridRefsElement);
				
				Map<Integer, TransientGridRef> transientGridRefs = component.getTransientGridRefs();
				List<TransientGridRef> transientGridRefList = new ArrayList<TransientGridRef>(
						transientGridRefs.values());
				Utils.sortByTransientGridRefId(transientGridRefList);
				
				for (TransientGridRef transientGridRef : transientGridRefList) {
					Element transientGridRefElement = doc.createElement(Dom.TRANSIENT_GRID_REF);
					transientGridRefsElement.appendChild(transientGridRefElement);
					
					int tgrId = transientGridRef.getId();
					transientGridRefElement.setAttribute(Dom.Attribute.ID, String.valueOf(tgrId));
					
					int transientRef = transientGridRef.getTransientRef();
					transientGridRefElement.setAttribute(Dom.Attribute.TRANSIENT_REF,
							String.valueOf(transientRef));
					
					int gridRef = transientGridRef.getGridRef();
					transientGridRefElement.setAttribute(Dom.Attribute.GRID_REF, String.valueOf(gridRef));
					
					int collectionSize = transientGridRef.getCollectionSize();
					transientGridRefElement.setAttribute(Dom.Attribute.COLLECTION_SIZE, String.valueOf(collectionSize));
				}
			}

			// <transients>
			Map<Integer, Transient> transients = coupling.getTransients();
			List<Integer> ids = new ArrayList<Integer>(transients.keySet());
			Collections.sort(ids);
			for (Integer id : ids) {
				Element transientElement = doc.createElement(Dom.TRANSIENT);
				transientElement.setAttribute(Dom.Attribute.ID, String.valueOf(id));
				transientElement.setAttribute(Dom.Attribute.TRANSIENT_STANDARD_NAME,
						transients.get(id).getTransientStandardName());
				transientsElement.appendChild(transientElement);
			}

			// <grid>
			Map<Integer, Grid> grids = coupling.getGrids();
			List<Integer> gridIds = new ArrayList<Integer>(grids.keySet());
			Collections.sort(gridIds);
			for (Integer id : gridIds) {
				Element gridElement = doc.createElement(Dom.GRID);
				gridElement.setAttribute(Dom.Attribute.ID, String.valueOf(id));
				String aliasName = grids.get(id).getAliasName();
				gridElement.setAttribute(Dom.Attribute.ALIAS_NAME, aliasName);
				gridsElement.appendChild(gridElement);
			}

			// <dates>
			Dates dates = coupling.getDates();
			Element startDateElement = doc.createElement(Dom.START_DATE);
			startDateElement.setTextContent(dates.getStartDate());
			datesElement.appendChild(startDateElement);
			Element endDateElement = doc.createElement(Dom.END_DATE);
			endDateElement.setTextContent(dates.getEndDate());
			datesElement.appendChild(endDateElement);
			Element calendarElement = doc.createElement(Dom.CALENDAR);
			calendarElement.setTextContent(String.valueOf(dates.getCalendar()));
			datesElement.appendChild(calendarElement);
			
			// <timestep_unit>
			String timestepUnit = coupling.getTimestepUnit().toString();
			timestepUnitElement.setTextContent(timestepUnit);

			// <couples>
			Map<String, Couple> couples = coupling.getCouples();
			List<String> coupleIds = new ArrayList<String>(couples.keySet());
			Collections.sort(coupleIds);
			for (String coupleId : coupleIds) {
				Couple couple = couples.get(coupleId);
				
				// <couple>
				Element coupleElement = doc.createElement(Dom.COUPLE);
				couplesElement.appendChild(coupleElement);
				
				// <component1>
				int component1Id = couple.getComponent1Id();
				Element component1Element = doc.createElement(Dom.COMPONENT1);
				component1Element.setAttribute(Dom.Attribute.COMPONENT_ID, String.valueOf(component1Id));
				coupleElement.appendChild(component1Element);
				
				// <component2>
				int component2Id = couple.getComponent2Id();
				Element component2Element = doc.createElement(Dom.COMPONENT2);
				component2Element.setAttribute(Dom.Attribute.COMPONENT_ID, String.valueOf(component2Id));
				coupleElement.appendChild(component2Element);
				
				// <transient_couple>
				for (TransientCouple transientCouple : couple.getTransientCouples()) {
					Element transientCoupleElement = doc.createElement(Dom.TRANSIENT_COUPLE);
					int transientId = transientCouple.getTransientId();
					transientCoupleElement.setAttribute(Dom.Attribute.TRANSIENT_ID,
							String.valueOf(transientId));
					
					// <source>
					Element sourceElement = doc.createElement(Dom.SOURCE);
					int sourceComponentRef = transientCouple.getSourceComponentRef();
					int sourceTransientGridRef = transientCouple.getSourceTransientGridRefId();
					sourceElement.setAttribute(Dom.Attribute.COMPONENT_REF,String.valueOf(sourceComponentRef));
					sourceElement.setAttribute(Dom.Attribute.TRANSIENT_GRID_REF,
							String.valueOf(sourceTransientGridRef));
					transientCoupleElement.appendChild(sourceElement);
					
					// <target>
					Element targetElement = doc.createElement(Dom.TARGET);
					int targetTransientGridRef = transientCouple.getTargetTransientGridRefIf();
					targetElement.setAttribute(Dom.Attribute.TRANSIENT_GRID_REF,
							String.valueOf(targetTransientGridRef));
					transientCoupleElement.appendChild(targetElement);
					
					// <timestep>
					Timestep timestep = transientCouple.getTimestep();
					Element timestepElement = doc.createElement(Dom.TIMESTEP);
					transientCoupleElement.appendChild(timestepElement);
					
					Element timestepSourceElement = doc.createElement(Dom.SOURCE);
					timestepSourceElement.setTextContent(timestep.getSource());
					timestepElement.appendChild(timestepSourceElement);
					
					Element timestepTargetElement = doc.createElement(Dom.TARGET);
					timestepTargetElement.setTextContent(timestep.getTarget());
					timestepElement.appendChild(timestepTargetElement);
					
					Element couplingPeriodElement = doc.createElement(Dom.COUPLING_PERIOD);
					couplingPeriodElement.setAttribute(Dom.Attribute.OPERATION, timestep.getCouplingOperation().toString());
					couplingPeriodElement.setTextContent(timestep.getCouplingPeriod());
					timestepElement.appendChild(couplingPeriodElement);

					Element sourceTimeLagElement = doc.createElement(Dom.SOURCE_TIMELAG);
					sourceTimeLagElement.setTextContent(String.valueOf(timestep.getSourceTimeLag()));
					//sourceTimeLagElement.set
					timestepElement.appendChild(sourceTimeLagElement);

					Element targetTimeLagElement = doc.createElement(Dom.TARGET_TIMELAG);
					targetTimeLagElement.setTextContent(String.valueOf(timestep.getTargetTimeLag()));
					timestepElement.appendChild(targetTimeLagElement);

					// <mapping_on_source>
					Element mappingOnSourceElement = doc.createElement(Dom.MAPPING_ON_SOURCE);
					boolean mappingOnSource = transientCouple.isMappingOnSource();
					mappingOnSourceElement.setTextContent(String.valueOf(transientCouple.getMappingOnSource()));
					transientCoupleElement.appendChild(mappingOnSourceElement);
					
					// <interpolation_requirements>
					Element interpolationRequirementsElement = doc.createElement(Dom.INTERPOLATION_REQUIREMENTS);
					transientCoupleElement.appendChild(interpolationRequirementsElement);
					
					List<Interpolation> interpolations = transientCouple.getInterpolations();
					
					if (interpolations != null) {
						for (Interpolation interpolation : interpolations) {
							
							Element interpolationElement = doc.createElement(Dom.INTERPOLATION);
							interpolationRequirementsElement.appendChild(interpolationElement);
							String method = interpolation.getMethod().getName();
							interpolationElement.setAttribute(Dom.Attribute.METHOD, method);
							
							for (int i = 0; i < interpolation.getAttributesNumber(); i++) {
								String name = interpolation.getAttributeName(i);
								String value = interpolation.getAttributeValue(i);
								interpolationElement.setAttribute(name, value);
							}
						}
					}
					
					//<enforce_write_weight_file>
					Element enforceWriteWeightFileElement = doc.createElement(Dom.ENFORCE_WRITE_WEIGHT_FILE);
					boolean enforceWriteWeightFile = transientCouple.isEnforceWriteWeightFile();
					enforceWriteWeightFileElement.setTextContent(String.valueOf(enforceWriteWeightFile));
					enforceWriteWeightFileElement.setAttribute(Dom.Attribute.FILENAME, transientCouple.getWeightFilename());
					transientCoupleElement.appendChild(enforceWriteWeightFileElement);

					coupleElement.appendChild(transientCoupleElement);
				}
			}
			
			Element createdElement = doc.createElement(Dom.CREATED);
			rootElement.appendChild(createdElement);
			createdElement.setAttribute(Dom.Attribute.DATE, new SimpleDateFormat(Default.CREATED_DATE_FORMAT).format(new Date()));
			createdElement.setAttribute(Dom.Attribute.TOOL, "YAC-CouplingGUI v." + VersionInfo.number);
			
			// write to xml file
			TransformerFactory transformerFactory = TransformerFactory.newInstance();
			Transformer transformer = transformerFactory.newTransformer();
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "3");
			
			DOMSource source = new DOMSource(doc);
			StreamResult result = new StreamResult(destination);
			transformer.transform(source, result);
			Logger.logInfo("File saved to " + destination.getPath());

		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (TransformerException e) {
			e.printStackTrace();
		}
	}
}
