/**
 * @file config_xml.c
 *
 * @copyright Copyright  (C)  2021 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *                                 Moritz Hanke <hanke@dkrz.de>
 *
 * @author Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 *         Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlschemas.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include "utils.h"
#include "config_xml.h"
#include "interp_method_avg.h"
#include "interp_method_conserv.h"
#include "interp_method_creep.h"
#include "interp_method_spmap.h"

struct name_type_pair {
  const char * name;
  int type;
};

static const struct name_type_pair interpolation_methods[] =
  {{.name = "average",               .type = AVERAGE},
   {.name = "n-nearest_neighbor",    .type = N_NEAREST_NEIGHBOR},
   {.name = "conservative",          .type = CONSERVATIVE},
   {.name = "source_to_target_map",  .type = SOURCE_TO_TARGET_MAP},
   {.name = "fixed_value",           .type = FIXED_VALUE},
   {.name = "user_file",             .type = USER_FILE},
   {.name = "check",                 .type = CHECK},
   {.name = "bernstein_bezier",      .type = BERNSTEIN_BEZIER},
   {.name = "radial_basis_function", .type = RADIAL_BASIS_FUNCTION},
   {.name = "creep",                 .type = CREEP}};
static const int num_interpolation_methods =
  sizeof(interpolation_methods)/sizeof(interpolation_methods[0]);

static const struct name_type_pair timestep_units[] =
  {{.name = "millisecond", .type = C_MILLISECOND},
   {.name = "second", .type = C_SECOND},
   {.name = "minute", .type = C_MINUTE},
   {.name = "hour", .type = C_HOUR},
   {.name = "day", .type = C_DAY},
   {.name = "month", .type = C_MONTH},
   {.name = "year", .type = C_YEAR},
   {.name = "ISO_format", .type = C_ISO_FORMAT}};
static const int num_timestep_units =
  sizeof(timestep_units)/sizeof(timestep_units[0]);

static const struct name_type_pair time_operations[] =
  {{.name = "accumulate", .type = ACCUMULATE},
   {.name = "average", .type = TIME_AVERAGE},
  {.name = "none", .type = INSTANT}};
static const int num_time_operations =
  sizeof(time_operations)/sizeof(time_operations[0]);

static const struct name_type_pair calendar_types[] =
  {{.name = "proleptic-gregorian", .type = PROLEPTIC_GREGORIAN},
   {.name = "360d", .type = YEAR_OF_360_DAYS},
   {.name = "365d", .type = YEAR_OF_365_DAYS}};
static const int num_calendar_types =
  sizeof(calendar_types)/sizeof(calendar_types[0]);

static char const * name_type_pair_get_name(
  struct name_type_pair const * pairs, size_t count, int type) {

  char const * name = NULL;
  for (size_t i = 0; (i < count) && (name == NULL); ++i)
    if (pairs[i].type == type) name = pairs[i].name;
  return name;
}

static int name_type_pair_get_type(
  struct name_type_pair const * pairs, size_t count, char const * name) {

  int type = INT_MAX;
  if (name != NULL)
    for (size_t i = 0; (i < count) && (type == INT_MAX); ++i)
      if (!strcmp(pairs[i].name, name)) type = pairs[i].type;
  return type;
}

static void check_is_numeric(const char * s, const char * routine_name) {
  char *p;
  YAC_ASSERT_F(
    (s != NULL) && (*s != '\0') && !isspace(*s) &&
    !isnan(strtod(s, &p)) && (*p == '\0'),
    "ERROR(%s): Value '%s' is not a numeric value.", routine_name, s);
}

static int get_node_attribute_int(
  xmlNodePtr node, char const * attribute_name, int line) {

  char const * attribute_value =
    (char const *)xmlGetProp(node, (xmlChar const *)attribute_name);

  check_is_numeric(attribute_value, "get_node_attribute_int");
  int attribute_value_int = atoi(attribute_value);
  xmlFree((void*)attribute_value);
  return attribute_value_int;
}

static char const * get_node_attribute_str(
  xmlNodePtr node, char const * attribute_name, int line) {

  char const * attribute_value =
    (char const *)xmlGetProp(node, (xmlChar const *)attribute_name);

  YAC_ASSERT_F(
    attribute_value != NULL,
    "ERROR attribute \"%s\" not found\n", attribute_name)

  char const * result = strdup((char const *)attribute_value);
  xmlFree((void*)attribute_value);
  return result;
}

static size_t get_number_of_elements_of_name(
  xmlNodePtr parent_node, const char * name) {

  YAC_ASSERT(parent_node != NULL, "ERROR xml parent node is NULL");

  xmlNodePtr child = parent_node->children;
  int n = 0;
  while (child) {
    if (!xmlStrcmp(child->name, (const xmlChar *) name)) ++n;
    child = child->next;
  }
  return (size_t)n;
}

static xmlNodePtr get_node_by_name(
  xmlNodePtr parent_node, const char * node_name, int child_is_optional) {

  YAC_ASSERT(parent_node != NULL, "ERROR xml parent node is NULL");

  xmlNodePtr node = parent_node->children;
  while (node) {
    if (!xmlStrcmp(node->name, (const xmlChar *) node_name)) break;
    node = node->next;
  }

  YAC_ASSERT_F(
    (node != NULL) || child_is_optional,
    "ERROR xml node \"%s\" is not available", node_name);

  return node;
}

static const char * get_date(xmlNodePtr dates_node, char const * node_name) {

  xmlNodePtr date_node = get_node_by_name(dates_node, node_name, 0);
  char const * date_node_content_str =
    (char const*)date_node->children->content;

  if (date_node_content_str == NULL) return NULL;

  YAC_ASSERT_F(
    strlen(date_node_content_str) < MAX_CALENDAR_STR_LEN,
    "ERROR \"%s\" string is too long", date_node_content_str)

  return strdup(date_node_content_str);
}

static int is_valid(const xmlDocPtr doc, const char * schema_filename) {

  xmlDocPtr schema_doc = xmlReadFile(schema_filename, NULL, XML_PARSE_NONET);
  YAC_ASSERT(
    schema_doc != NULL, "Schema cannot be loaded or is not well-formed")

  xmlSchemaParserCtxtPtr parser_ctxt = xmlSchemaNewDocParserCtxt(schema_doc);
  YAC_ASSERT(
    parser_ctxt != NULL, "Unable to create parser context for schema")

  xmlSchemaPtr schema = xmlSchemaParse(parser_ctxt);
  YAC_ASSERT(schema != NULL, "Schema is not valid")

  xmlSchemaValidCtxtPtr valid_ctxt = xmlSchemaNewValidCtxt(schema);
  YAC_ASSERT(
    valid_ctxt != NULL, "Unable to create validation context for schema")

  int is_valid = xmlSchemaValidateDoc(valid_ctxt, doc) == 0;

  xmlSchemaFreeValidCtxt(valid_ctxt);
  xmlSchemaFree(schema);
  xmlSchemaFreeParserCtxt(parser_ctxt);
  xmlFreeDoc(schema_doc);

  return is_valid;
}

static enum yac_time_unit_type get_timestep_unit(xmlNodePtr coupling) {

  xmlNodePtr timestep_unit_node = get_node_by_name(coupling, "timestep_unit", 0);
  char * timestep_unit_str = (char *) timestep_unit_node->children->content;

  int timestep_unit =
    name_type_pair_get_type(
      timestep_units, num_timestep_units, timestep_unit_str);

  YAC_ASSERT_F(
    timestep_unit != INT_MAX, "Unsupported time unit '%s'.", timestep_unit_str)

  return (enum yac_time_unit_type)timestep_unit;
}

static void get_grids(
  xmlNodePtr grids_node, struct yac_couple_config_grid * grids) {

  xmlNodePtr children = grids_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"grid")) {

      xmlNodePtr grid_node = children;
      grids->name = get_node_attribute_str(grid_node, "alias_name", __LINE__);
      grids->id = get_node_attribute_int(grid_node, "id", __LINE__);

      YAC_ASSERT(
        strlen(grids->name) < MAX_FILE_NAME_LENGTH,
        "ERROR grid name is too long")

      ++grids;
    }
    children = children->next;
  }
}

static void get_transients(
  const xmlNodePtr transients_node,
  struct yac_couple_config_transient * transients) {

  xmlNodePtr children = transients_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"transient")) {

      xmlNodePtr transient_node = children;
      transients->name =
        get_node_attribute_str(
          transient_node, "transient_standard_name", __LINE__);
      transients->id = get_node_attribute_int(transient_node, "id", __LINE__);
      transients->collection_size = SIZE_MAX;
      ++transients;
    }
    children = children->next;
  }
}

static size_t get_transient_idx(
  size_t num_transients, struct yac_couple_config_transient * transients,
  int transient_id) {

  size_t transient_idx = SIZE_MAX;
  for (size_t i = 0; (i < num_transients) && (transient_idx == SIZE_MAX); ++i)
    if (transients[i].id == transient_id) transient_idx = i;
  return transient_idx;
}

static size_t get_grid_idx(
  size_t num_grids, struct yac_couple_config_grid * grids,
  int grid_id) {

  size_t grid_idx = SIZE_MAX;
  for (size_t i = 0; (i < num_grids) && (grid_idx == SIZE_MAX); ++i)
    if (grids[i].id == grid_id) grid_idx = i;
  return grid_idx;
}

static void get_component_transients(
  xmlNodePtr transient_grid_refs_node,
  size_t num_transients, struct yac_couple_config_transient * transients,
  size_t num_grids, struct yac_couple_config_grid * grids,
  int * transient_grid_ref_ids,
  struct yac_couple_config_component_transient * component_transients) {

  xmlNodePtr children = transient_grid_refs_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"transient_grid_ref")) {

      xmlNodePtr transient_grid_ref_node = children;
      int id =
        get_node_attribute_int(
          transient_grid_ref_node, "id", __LINE__);
      int transient_ref =
        get_node_attribute_int(
          transient_grid_ref_node, "transient_ref", __LINE__);
      int grid_ref =
        get_node_attribute_int(
          transient_grid_ref_node, "grid_ref", __LINE__);
      int collection_size =
        (size_t)get_node_attribute_int(
          transient_grid_ref_node, "collection_size", __LINE__);

      size_t transient_idx =
        get_transient_idx(num_transients, transients, transient_ref);
      YAC_ASSERT(
        transient_idx != SIZE_MAX,
        "ERROR could not find transient id from transient grid ref in "
        "list of transients")

      size_t grid_idx = get_grid_idx(num_grids, grids, grid_ref);
      YAC_ASSERT(
        grid_idx != SIZE_MAX,
        "ERROR could not find grid id from grid grid ref in list of grids")

      if (transients[transient_idx].collection_size == SIZE_MAX)
        transients[transient_idx].collection_size = collection_size;
      else
        YAC_ASSERT(
          transients[transient_idx].collection_size == collection_size,
          "ERROR(get_component_transients): collection size missmatch")

      *transient_grid_ref_ids = id;
      component_transients->transient_idx = transient_idx;
      component_transients->grid_idx = grid_idx;
      component_transients++;
      transient_grid_ref_ids++;
    }
    children = children->next;
  }
}

static void get_components(
  xmlNodePtr components_node,
  size_t num_transients, struct yac_couple_config_transient * transients,
  size_t num_grids, struct yac_couple_config_grid * grids,
  int ** transient_grid_ref_ids,
  struct yac_couple_config_component * components) {

  xmlNodePtr children = components_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"component")) {

      xmlNodePtr component_node = children;
      xmlNodePtr transient_grid_refs_node =
        get_node_by_name(component_node, "transient_grid_refs", 1);
      size_t num_component_transients =
        (transient_grid_refs_node != NULL)?
          get_number_of_elements_of_name(
            transient_grid_refs_node, "transient_grid_ref"):0;
      *transient_grid_ref_ids =
        xmalloc(num_component_transients * sizeof(**transient_grid_ref_ids));
      struct yac_couple_config_component_transient * component_transients =
        xmalloc(num_component_transients * sizeof(*component_transients));
      if (num_component_transients > 0)
        get_component_transients(
          transient_grid_refs_node, num_transients, transients,
          num_grids, grids, *transient_grid_ref_ids, component_transients);

      xmlNodePtr name_node = get_node_by_name(component_node, "name", 0);

      components->component_transients = component_transients;
      components->num_component_transients = num_component_transients;
      components->name = strdup((char const*)name_node->children->content);
      components->id =
        get_node_attribute_int(component_node, "id", __LINE__);
      ++transient_grid_ref_ids;
      ++components;
    }
    children = children->next;
  }
}

static enum yac_reduction_type get_reduction_type(
  xmlNodePtr coupling_period_node) {

  char * operation_str =
    (char *)xmlGetProp(coupling_period_node, (const xmlChar *)"operation");
  YAC_ASSERT(
    operation_str != NULL,
    "ERROR(get_reduction_type): no \"operation\"-string provided")

  int reduction_type =
    name_type_pair_get_type(
      time_operations, num_time_operations, operation_str);

  YAC_ASSERT_F(
    reduction_type != INT_MAX,
    "ERROR(get_reduction_type): '%s' is not a valid time reduction operation",
    operation_str)
  xmlFree((void*)operation_str);

  return (enum yac_reduction_type)reduction_type;
}

static enum yac_interpolation_list get_interpolation_method(
  xmlNodePtr interpolation_node) {

  char const * method_name =
    (char const *)xmlGetProp(interpolation_node, (const xmlChar *)"method");
  YAC_ASSERT(
    method_name != NULL,
    "ERROR(get_interpolation_method): no \"method\"-string provided")

  int interpolation_method =
    name_type_pair_get_type(
      interpolation_methods, num_interpolation_methods, method_name);
  YAC_ASSERT_F(
    interpolation_method != INT_MAX,
    "Unsupported interpolation method '%s'", method_name)
  xmlFree((void*)method_name);

  return (enum yac_interpolation_list)interpolation_method;
}

static int get_interpolation_parameter_value_int(
  xmlNodePtr interpolation_node, const char * parameter, int default_value) {

  const char * value =
    (const char *)xmlGetProp(interpolation_node, (const xmlChar *)parameter);

  if (value == NULL) return default_value;

  check_is_numeric(value, "get_interpolation_parameter_value_int");
  int result = atoi(value);
  xmlFree((void*)value);
  return result;
}

static double get_interpolation_parameter_value_double(
  xmlNodePtr interpolation_node, const char * parameter, double default_value) {

  const char * value =
    (const char *)xmlGetProp(interpolation_node, (const xmlChar *)parameter);

  if (value == NULL) return default_value;

  check_is_numeric(value, "get_interpolation_parameter_value_double");
  double result = atof(value);
  xmlFree((void*)value);
  return result;
}

static int get_interpolation_parameter_value_type(
  xmlNodePtr interpolation_node, const char * parameter, int default_value,
  struct name_type_pair * pairs, size_t pair_count,
  const char * routine, enum yac_interpolation_list method) {

  char * value =
    (char *)xmlGetProp(interpolation_node, (const xmlChar *)parameter);

  if (value == NULL) return default_value;

  // convert to lower case
  for(int i = 0; value[i]; i++) value[i] = tolower(value[i]);

  int type = name_type_pair_get_type(pairs, pair_count, value);
  YAC_ASSERT_F(
    type != INT_MAX,
    "ERROR(%s): '%s' is not a invalid \"%s\" value for %s",
    routine, value, parameter,
    name_type_pair_get_name(
      interpolation_methods, num_interpolation_methods, (int)method))

  xmlFree((void*)value);
  return type;
}

static int bool_str_2_int(const char * bool_str) {

  int ret =
    name_type_pair_get_type(
      (struct name_type_pair[])
        {{.name = "true",  .type = 1},
         {.name = "false", .type = 0}}, 2, bool_str);
  YAC_ASSERT_F(
    ret != INT_MAX,
    "ERROR(bool_str_2_int): '%s' is not a invalid boolean string", bool_str)

  return ret;
}

static int get_interpolation_parameter_value_bool(
  xmlNodePtr interpolation_node, const char * parameter, int default_value) {

  const char * value =
    (const char *)xmlGetProp(interpolation_node, (const xmlChar *)parameter);

  if (value == NULL) return default_value;

  int result = bool_str_2_int(value);
  xmlFree((void*)value);
  return result;
}

static struct yac_interp_stack_config get_interp_stack(
  xmlNodePtr interpolation_requirements_node,
  char const * src_grid_name, char const * tgt_grid_name) {

  struct yac_interp_stack_config interp_stack;

  interp_stack.size =
    get_number_of_elements_of_name(
      interpolation_requirements_node, "interpolation");
  union yac_interp_stack_config_entry * config =
    ((interp_stack.config =
        xcalloc(interp_stack.size, sizeof(*(interp_stack.config)))));

  xmlNodePtr children = interpolation_requirements_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"interpolation")) {

      xmlNodePtr interpolation_node = children;
      config->general.type = get_interpolation_method(interpolation_node);

      switch ((int)(config->general.type)) {

        default:
        case (AVERAGE): {

          config->average.partial_coverage =
            get_interpolation_parameter_value_bool(
              interpolation_node, "partial_coverage", 0);
          config->average.reduction_type =
            (enum yac_interp_avg_weight_type)
              get_interpolation_parameter_value_type(
                interpolation_node, "weighted", AVG,
                (struct name_type_pair[])
                  {{.name = "distance_weighted", .type = DIST},
                   {.name = "arithmetic_average", .type = AVG},
                   {.name = "barycentric_coordinate", .type = BARY}},
                3, "get_interp_stack", AVERAGE);
          break;
        }
        case (N_NEAREST_NEIGHBOR): {

          int nn =
            get_interpolation_parameter_value_int(
              interpolation_node, "n", -1);

          YAC_ASSERT(
            nn != -1,
            "ERROR(get_interp_stack): N_NEAREST_NEIGHBOR requires valid n!")

          enum yac_interp_nnn_weight_type reduction_type =
            (enum yac_interp_nnn_weight_type)
              get_interpolation_parameter_value_type(
                interpolation_node, "weighted", NNN_AVG,
                (struct name_type_pair[])
                  {{.name = "distance_weighted", .type = NNN_DIST},
                   {.name = "gauss_weighted", .type = NNN_GAUSS},
                   {.name = "arithmetic_average", .type = NNN_AVG}},
                3, "get_interp_stack", N_NEAREST_NEIGHBOR);

          double gauss_scale = -1;

          if (reduction_type == NNN_GAUSS)
            gauss_scale =
              get_interpolation_parameter_value_double(
                interpolation_node, "gauss_scale", YAC_DEFAULT_GAUSS_SCALE);

          config->n_nearest_neighbor.config.data.gauss_scale = gauss_scale;
          config->n_nearest_neighbor.config.n = (size_t)nn;
          config->n_nearest_neighbor.config.type = reduction_type;
          break;
        }
        case (CONSERVATIVE): {

          int order =
            get_interpolation_parameter_value_int(
              interpolation_node, "order", 1);

          YAC_ASSERT(
            (order == 1) || (order == 2),
            "ERROR(get_interp_stack): invalid value for conservative "
            "remapping XML parameter order")

          int enforced_conserv =
            get_interpolation_parameter_value_bool(
              interpolation_node, "enforced_conservation", 0);
          int partial_coverage =
            get_interpolation_parameter_value_bool(
              interpolation_node, "partial_coverage", 0);
          enum yac_interp_method_conserv_normalisation normalisation =
            (enum yac_interp_method_conserv_normalisation)
              get_interpolation_parameter_value_type(
                interpolation_node, "normalisation", CONSERV_DESTAREA,
                (struct name_type_pair[])
                  {{.name = "fracarea", .type = CONSERV_FRACAREA},
                   {.name = "destarea", .type = CONSERV_DESTAREA}},
                2, "get_interp_stack", CONSERVATIVE);

          config->conservative.order = order;
          config->conservative.enforced_conserv = enforced_conserv;
          config->conservative.partial_coverage = partial_coverage;
          config->conservative.normalisation = normalisation;
          break;
        }
        case (SOURCE_TO_TARGET_MAP):
        {

          double spread_distance =
            get_interpolation_parameter_value_double(
              interpolation_node, "spread_distance",
              YAC_DEFAULT_SPREAD_DISTANCE);

          YAC_ASSERT(
            (spread_distance >= 0.0) && (spread_distance < 90.0),
            "ERROR(get_interp_stack): invalid spread_distance "
            "(has to be >= 0 and < 90)")

          double max_search_distance =
            get_interpolation_parameter_value_double(
              interpolation_node, "max_search_distance",
              YAC_DEFAULT_MAX_SEARCH_DISTANCE);

          YAC_ASSERT(
            (max_search_distance >= 0.0) && (max_search_distance < 180.0),
            "ERROR(get_interp_stack): invalid max_search_distance "
            "(has to be >= 0 and < 180)")

          enum yac_interp_spmap_weight_type weight_type =
            (enum yac_interp_spmap_weight_type)
              get_interpolation_parameter_value_type(
                interpolation_node, "weighted", SPMAP_AVG,
                (struct name_type_pair[])
                  {{.name = "distance_weighted", .type = SPMAP_DIST},
                   {.name = "arithmetic_average", .type = SPMAP_AVG}},
                2, "get_interp_stack", SOURCE_TO_TARGET_MAP);

          config->spmap.spread_distance = spread_distance * YAC_RAD;
          config->spmap.max_search_distance = max_search_distance * YAC_RAD;
          config->spmap.weight_type = weight_type;
          break;
        }
        case (FIXED_VALUE):
        {
          double fixed_value =
            get_interpolation_parameter_value_double(
              interpolation_node, "user_value", DBL_MAX);

          YAC_ASSERT(
            fixed_value != DBL_MAX,
            "ERROR(get_interp_stack): start_search: FIXED_VALUE "
            "requires a value defined by the user!")

          config->fixed.value = fixed_value;
          break;
        }
        case (USER_FILE):
        {

          // get the filename containing the weights

          char const * filename =
            (const char *)xmlGetProp(
              interpolation_node, (const xmlChar *)"filename");
          YAC_ASSERT(
            filename != NULL,
            "ERROR(get_interp_stack): "
            "USER_FILE requires a filename containing "
            "weights defined by the user!")

          YAC_ASSERT(
            strlen(filename) < MAX_FILE_NAME_LENGTH,
            "ERROR(get_interp_stack): USER_FILE filename is too long!")
          strcpy(config->user_file.filename, filename);

          xmlFree((void*)filename);

          config->user_file.io_proc_step = -1;
          strcpy(config->user_file.src_grid_name, src_grid_name);
          strcpy(config->user_file.tgt_grid_name, tgt_grid_name);

          break;
        }
        case (CHECK):
        {
          char const * constructor_key =
            (const char *)xmlGetProp(
              interpolation_node, (const xmlChar *)"constructor_key");

          if (constructor_key != NULL) {

            YAC_ASSERT(
              strlen(constructor_key) < MAX_ROUTINE_NAME_LENGTH,
              "ERROR(get_interp_stack): CHECK constructor_key is too long!")
            strcpy(config->check.constructor_key, constructor_key);

            xmlFree((void*)constructor_key);
          }

          char const * do_search_key =
            (const char *)xmlGetProp(
              interpolation_node, (const xmlChar *)"do_search_key");

          if (do_search_key != NULL) {

            YAC_ASSERT(
              strlen(do_search_key) < MAX_ROUTINE_NAME_LENGTH,
              "ERROR(get_interp_stack): CHECK do_search_key is too long!")
            strcpy(config->check.do_search_key, do_search_key);

            xmlFree((void*)do_search_key);
          }

          break;
        }
        case (BERNSTEIN_BEZIER):
        {
          break;
        }
        case (RADIAL_BASIS_FUNCTION):
        {
          int rbf_kernel_type =
            get_interpolation_parameter_value_type(
              interpolation_node, "rbf_kernel", 0,
              (struct name_type_pair[])
                {{.name = "gauss_kernel", .type = 0}},
              1, "get_interp_stack", RADIAL_BASIS_FUNCTION);
          UNUSED(rbf_kernel_type);

          config->radial_basis_function.config.data.rbf_scale =
              get_interpolation_parameter_value_double(
                interpolation_node, "rbf_scale", YAC_DEFAULT_RBF_SCALE);
          config->radial_basis_function.config.n =
            get_interpolation_parameter_value_int(interpolation_node, "n", 9);
          config->n_nearest_neighbor.config.type = NNN_RBF;
          break;
        }
        case (CREEP):
        {

          config->creep.creep_distance =
            get_interpolation_parameter_value_int(
              interpolation_node, "creep_distance",
              YAC_DEFAULT_CREEP_DISTANCE);
          break;
        }
      }
      ++config;
    }
    children = children->next;
  }

  return interp_stack;
}

static void get_transient_couples(
  xmlNodePtr couple_node,
  size_t num_transients, struct yac_couple_config_transient * transients,
  size_t num_grids, struct yac_couple_config_grid * grids,
  size_t num_components, struct yac_couple_config_component * components,
  size_t * component_indices, int ** transient_grid_ref_ids,
  struct yac_couple_config_transient_couple * transient_couples) {

  xmlNodePtr children = couple_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"transient_couple")) {

      xmlNodePtr transient_couple_node = children;

      size_t transient_idx =
        get_transient_idx(
          num_transients, transients,
          get_node_attribute_int(
            transient_couple_node, "transient_id", __LINE__));
      YAC_ASSERT(
        transient_idx != SIZE_MAX,
        "ERROR could not find transient id from transient couple in "
        "list of transients")

      xmlNodePtr source_node =
        get_node_by_name(transient_couple_node, "source", 0);
      int source_component_ref =
        get_node_attribute_int(source_node, "component_ref", __LINE__);
      YAC_ASSERT(
        (source_component_ref == 1) || (source_component_ref == 2),
        "ERROR invalid component_ref in transient_couple "
        "(has to be either 1 or 2)")
      size_t source_component_idx =
        component_indices[source_component_ref - 1];
      int source_transient_grid_ref =
        get_node_attribute_int(source_node, "transient_grid_ref", __LINE__);
      size_t source_component_transient_idx = SIZE_MAX;
      for (size_t i = 0;
           (source_component_transient_idx == SIZE_MAX) &&
           (i < components[source_component_idx].num_component_transients); ++i)
        if (transient_grid_ref_ids[source_component_idx][i] ==
            source_transient_grid_ref) source_component_transient_idx = i;
      YAC_ASSERT_F(
        source_component_transient_idx != SIZE_MAX,
        "ERROR(get_transient_couples): could not find source "
        "transient_grid_ref_id %d in list of transient_grid_refs of "
        "source component %s (id %d)",
        (int)source_transient_grid_ref, components[source_component_idx].name,
        (int)(components[source_component_idx].id));

      xmlNodePtr target_node =
        get_node_by_name(transient_couple_node, "target", 0);
      size_t target_component_idx =
        component_indices[(source_component_ref - 1) ^ 1];
      int target_transient_grid_ref =
        get_node_attribute_int(target_node, "transient_grid_ref", __LINE__);
      size_t target_component_transient_idx = SIZE_MAX;
      for (size_t i = 0;
           (target_component_transient_idx == SIZE_MAX) &&
           (i < components[target_component_idx].num_component_transients); ++i)
        if (transient_grid_ref_ids[target_component_idx][i] ==
            target_transient_grid_ref) target_component_transient_idx = i;
      YAC_ASSERT_F(
        target_component_transient_idx != SIZE_MAX,
        "ERROR(get_transient_couples): could not find target "
        "transient_grid_ref_id %d in list of transient_grid_refs of "
        "target component \"%s\" (id %d)",
        (int)target_transient_grid_ref, components[target_component_idx].name,
        (int)(components[target_component_idx].id));

      xmlNodePtr timestep_node =
        get_node_by_name(transient_couple_node, "timestep", 0);
      xmlNodePtr coupling_period_node =
        get_node_by_name(timestep_node, "coupling_period", 0);
      char const * coupling_period =
        strdup((char *)coupling_period_node->children->content);
      enum yac_reduction_type coupling_period_operation =
        get_reduction_type(coupling_period_node);

      xmlNodePtr source_timestep_node =
        get_node_by_name(timestep_node, "source", 0);
      char const * source_timestep =
        strdup((char *)source_timestep_node->children->content);
      xmlNodePtr source_timelag_node =
        get_node_by_name(timestep_node, "source_timelag", 0);
      char * source_lag_value = (char*)source_timelag_node->children->content;
      check_is_numeric(source_lag_value, "get_transient_couples");
      int source_lag = atoi(source_lag_value);

      xmlNodePtr target_timestep_node =
        get_node_by_name(timestep_node, "target", 0);
      char const * target_timestep =
        strdup((char *)target_timestep_node->children->content);
      xmlNodePtr target_timelag_node =
        get_node_by_name(timestep_node, "target_timelag", 0);
      char * target_lag_value = (char*)target_timelag_node->children->content;
      check_is_numeric(target_lag_value, "get_transient_couples");
      int target_lag = atoi(target_lag_value);

      xmlNodePtr interpolation_requirements_node =
        get_node_by_name(
          transient_couple_node, "interpolation_requirements", 0);
      struct yac_interp_stack_config interp_stack =
        get_interp_stack(
          interpolation_requirements_node,
          grids[
            components[source_component_idx].
              component_transients[source_component_transient_idx].
                grid_idx].name,
          grids[
            components[target_component_idx].
              component_transients[target_component_transient_idx].
                grid_idx].name);

      xmlNodePtr mapping_on_source_node =
        get_node_by_name(transient_couple_node, "mapping_on_source", 1);
      int mapping_on_source =
        (mapping_on_source_node == NULL) ||
         bool_str_2_int(
           (char const *)(mapping_on_source_node->children->content));

      xmlNodePtr enforce_write_weight_file_node =
        get_node_by_name(
          transient_couple_node, "enforce_write_weight_file", 1);
      int enforce_write_weight_file =
        (enforce_write_weight_file_node != NULL)?
          bool_str_2_int(
            (char const *)(
              enforce_write_weight_file_node->children->content)):0;

      char const * weight_file_name =
        (enforce_write_weight_file_node != NULL)?
          (char *)xmlGetProp(
            enforce_write_weight_file_node, (const xmlChar *)"filename"):NULL;
      YAC_ASSERT(
        !enforce_write_weight_file || (weight_file_name != NULL),
        "Attribute \"filename\" not found.")

      transient_couples->source.component_idx = source_component_idx;
      transient_couples->source.component_transient_idx =
        source_component_transient_idx;
      transient_couples->source.timestep = source_timestep;
      transient_couples->source.lag = source_lag;
      transient_couples->target.component_idx = target_component_idx;
      transient_couples->target.component_transient_idx =
        target_component_transient_idx;
      transient_couples->target.timestep = target_timestep;
      transient_couples->target.lag = target_lag;
      transient_couples->mapping_on_source = mapping_on_source;
      transient_couples->interp_stack = interp_stack;
      transient_couples->coupling_period_operation = coupling_period_operation;
      transient_couples->coupling_period = coupling_period;
      transient_couples->enforce_write_weight_file = enforce_write_weight_file;
      transient_couples->weight_file_name = weight_file_name;
      ++transient_couples;
    }
    children = children->next;
  }
}

static size_t get_component_idx(
  size_t num_components, struct yac_couple_config_component * components,
  int component_id) {

  size_t component_idx = SIZE_MAX;
  for (size_t i = 0; (i < num_components) && (component_idx == SIZE_MAX); ++i)
    if (components[i].id == component_id) component_idx = i;
  return component_idx;
}

static void get_couples(
  xmlNodePtr couples_node,
  size_t num_transients, struct yac_couple_config_transient * transients,
  size_t num_grids, struct yac_couple_config_grid * grids,
  size_t num_components, struct yac_couple_config_component * components,
  int ** transient_grid_ref_ids, struct yac_couple_config_couple * couples) {

  xmlNodePtr children = couples_node->children;
  while (children) {
    if (!xmlStrcmp(children->name, (const xmlChar *)"couple")) {

      xmlNodePtr couple_node = children;

      size_t component_indices[2] = {
        get_component_idx(
          num_components, components,
          get_node_attribute_int(
            get_node_by_name(couple_node, "component1", 0),
            "component_id", __LINE__)),
        get_component_idx(
          num_components, components,
          get_node_attribute_int(
            get_node_by_name(couple_node, "component2", 0),
            "component_id", __LINE__))};

      YAC_ASSERT(
        (component_indices[0] != SIZE_MAX) &&
        (component_indices[1] != SIZE_MAX),
        "ERROR component id of couple not in list of components")

      size_t num_transient_couples =
        get_number_of_elements_of_name(couple_node, "transient_couple");
      struct yac_couple_config_transient_couple * transient_couples =
        xmalloc(num_transient_couples * sizeof(*transient_couples));
      get_transient_couples(
        couple_node, num_transients, transients, num_grids, grids,
        num_components, components, component_indices, transient_grid_ref_ids,
        transient_couples);

      couples->component_indices[0] = component_indices[0];
      couples->component_indices[1] = component_indices[1];
      couples->transient_couples = transient_couples;
      couples->num_transient_couples = num_transient_couples;
      ++couples;
    }
    children = children->next;
  };
}

static int redirect_stdout(xmlNodePtr coupling) {

  xmlNodePtr redirect_node = get_node_by_name(coupling, "redirect", 1);

  if (redirect_node == NULL) return 0;

  char * value =
    (char *) xmlGetProp(redirect_node, (const xmlChar *)"redirect_stdout");

  int redirect = bool_str_2_int(value);
  xmlFree(value);

  if (redirect) {

    value =
      (char *) xmlGetProp(redirect_node, (const xmlChar *)"redirect_of_root");

    // increment only if redirect of all processes is requested
    redirect += !bool_str_2_int(value);
    xmlFree(value);
  }


  return redirect;
}

static calendarType get_calendar(xmlNodePtr dates_node) {

  xmlNodePtr calendar_node = get_node_by_name(dates_node, "calendar", 0);
  char * calendar_str = (char *) calendar_node->children->content;

  int calendar_type =
    name_type_pair_get_type(calendar_types, num_calendar_types, calendar_str);

  return
    (calendarType)
      ((calendar_type != INT_MAX)?(calendar_type):(CALENDAR_NOT_SET));
}

struct yac_couple_config * yac_couple_config_xml_new(
  const char * xml_filename, const char * schema_filename) {

  // error if file does not exist
  YAC_ASSERT_F(
    yac_file_exists(xml_filename), "File %s could not be found.", xml_filename)

  // parse the XML file
  xmlDocPtr document = xmlReadFile(xml_filename, "UTF-8", 0);

  // error if parsing was not successful
  YAC_ASSERT_F(
    document != NULL, "Could not parse file %s", xml_filename)

  // validate XML against schema
  YAC_ASSERT(
    is_valid(document, schema_filename),
    "XML does not validate against schema.")

  // get root element
  xmlNodePtr coupling_node = xmlDocGetRootElement(document);

  // error if file was empty
  YAC_ASSERT_F(
    coupling_node != NULL, "File %s is empty.", xml_filename)

  struct yac_couple_config * couple_config = xmalloc(1 * sizeof(*couple_config));

  xmlNodePtr dates_node = get_node_by_name(coupling_node, "dates", 0);
  couple_config->start_datetime = get_date(dates_node, "start_date");
  couple_config->end_datetime = get_date(dates_node, "end_date");
  couple_config->calendar = get_calendar(dates_node);
  couple_config->timestep_unit = get_timestep_unit(coupling_node);

  xmlNodePtr grids_node = get_node_by_name(coupling_node, "grids", 0);
  size_t num_grids =
    ((couple_config->num_grids =
        get_number_of_elements_of_name(grids_node, "grid")));
  struct yac_couple_config_grid * grids =
    ((couple_config->grids = xmalloc(num_grids * sizeof(*grids))));
  get_grids(grids_node, grids);

  xmlNodePtr transients_node =
    get_node_by_name(coupling_node, "transients", 0);
  size_t num_transients =
    ((couple_config->num_transients =
        get_number_of_elements_of_name(transients_node, "transient")));
  struct yac_couple_config_transient * transients =
    ((couple_config->transients =
        xmalloc(num_transients * sizeof(*transients))));
  get_transients(transients_node, transients);

  xmlNodePtr components_node =
    get_node_by_name(coupling_node, "components", 0);
  size_t num_components =
    ((couple_config->num_components =
        get_number_of_elements_of_name(components_node, "component")));
  struct yac_couple_config_component * components =
    ((couple_config->components =
        xmalloc(num_components * sizeof(*components))));
  int ** transient_grid_ref_ids =
    xmalloc(num_components * sizeof(*transient_grid_ref_ids));
  get_components(
    components_node, num_transients, transients, num_grids, grids,
    transient_grid_ref_ids, components);

  xmlNodePtr couples_node = get_node_by_name(coupling_node, "couples", 0);
  size_t num_couples =
    ((couple_config->num_couples =
        get_number_of_elements_of_name(couples_node, "couple")));
  struct yac_couple_config_couple * couples =
    ((couple_config->couples = xmalloc(num_couples * sizeof(*couples))));
  get_couples(
    couples_node, num_transients, transients, num_grids, grids,
    num_components, components, transient_grid_ref_ids, couples);

  for (size_t component_idx = 0; component_idx < num_components;
       ++component_idx)
    free(transient_grid_ref_ids[component_idx]);
  free(transient_grid_ref_ids);

  couple_config->redirect_stdout = redirect_stdout(coupling_node);

  // destroy parser
  xmlCleanupParser();
  xmlFreeDoc(document);

  return couple_config;
}
