/**
 * @file yac_mpi.c
 *
 * copied from yaxt
 *
 * @copyright Copyright  (C)  2020 Moritz Hanke <hanke@dkrz.de>
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <inttypes.h>
#include <limits.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mpi.h>
#include "yac_mpi.h"
#include "geometry.h"
#include "ensure_array_size.h"
#include "core/core.h"

static int mpi_initialised_by_yac = 0;
static int yaxt_initialised_by_yac = 0;
static int * comm_buffer = NULL;
static size_t comm_buffer_array_size = 0;
static int comm_buffer_in_use = 0;
static int init_count = 0;

//taken from http://beige.ucs.indiana.edu/I590/node85.html
void yac_mpi_error(int error_code, MPI_Comm comm) {
  int rank;
  MPI_Comm_rank(comm, &rank);

  char error_string[MPI_MAX_ERROR_STRING];
  int length_of_error_string, error_class;

  MPI_Error_class(error_code, &error_class);
  MPI_Error_string(error_class, error_string, &length_of_error_string);
  fprintf(stderr, "%3d: %s\n", rank, error_string);
  MPI_Abort(comm, error_code);
}

void yac_alltoallv_p2p(
  void const * send_buffer, int const * sendcounts, int const * sdispls,
  void * recv_buffer, int const * recvcounts, int const * rdispls,
  size_t dt_size, MPI_Datatype dt, MPI_Comm comm) {

#define USE_P2P_ALLTOALLV
#ifdef USE_P2P_ALLTOALLV
  int comm_rank, comm_size;
  yac_mpi_call(MPI_Comm_rank(comm, &comm_rank), comm);
  yac_mpi_call(MPI_Comm_size(comm, &comm_size), comm);

  int req_count = 0;
  for (int i = 0; i < comm_size; ++i)
    req_count += (sendcounts[i] > 0) + (recvcounts[i] > 0);
  MPI_Request * req = xmalloc((size_t)req_count * sizeof(*req));

  req_count = 0;
  for (int j = 0, lb = comm_rank, ub = comm_size; j < 2;
       ++j, lb = 0, ub = comm_rank) {
    for (int i = lb; i < ub; ++i) {
      if (sendcounts[i] > 0) {

        yac_mpi_call(
          MPI_Isend(
            (void const *)((unsigned char *)send_buffer +
                           dt_size * (size_t)(sdispls[i])),
            sendcounts[i], dt, i, 0,
            comm, req + req_count), comm);
        ++req_count;
      }
      if (recvcounts[i] > 0) {
        yac_mpi_call(
          MPI_Irecv(
            (void *)((unsigned char *)recv_buffer +
                     dt_size * (size_t)(rdispls[i])),
            recvcounts[i], dt, i, 0,
            comm, req + req_count), comm);
        ++req_count;
      }
    }
  }
  yac_mpi_call(MPI_Waitall(req_count, req, MPI_STATUSES_IGNORE), comm);
  free(req);
#else // USE_P2P_ALLTOALLV
  yac_mpi_call(
    MPI_Alltoallv(send_buffer, sendcounts, sdispls, dt,
                  recv_buffer, recvcounts, rdispls, dt, comm), comm);
#endif // USE_P2P_ALLTOALLV
}

void yac_alltoallv_int_p2p(
  int const * send_buffer, int const * sendcounts, int const * sdispls,
  int * recv_buffer, int const * recvcounts, int const * rdispls,
  MPI_Comm comm) {

  yac_alltoallv_p2p(
    (void const *)send_buffer, sendcounts, sdispls,
    (void *)recv_buffer, recvcounts, rdispls,
    sizeof(int), MPI_INT, comm);
}

void yac_alltoallv_yac_int_p2p(
  yac_int const * send_buffer, int const * sendcounts, int const * sdispls,
  yac_int * recv_buffer, int const * recvcounts, int const * rdispls,
  MPI_Comm comm) {

  yac_alltoallv_p2p(
    (void const *)send_buffer, sendcounts, sdispls,
    (void *)recv_buffer, recvcounts, rdispls,
    sizeof(yac_int), Xt_int_dt, comm);
}

void yac_alltoallv_uint64_p2p(
  uint64_t const * send_buffer, int const * sendcounts, int const * sdispls,
  uint64_t * recv_buffer, int const * recvcounts, int const * rdispls,
  MPI_Comm comm) {

  yac_alltoallv_p2p(
    (void const *)send_buffer, sendcounts, sdispls,
    (void *)recv_buffer, recvcounts, rdispls,
    sizeof(uint64_t), MPI_UINT64_T, comm);
}

void yac_alltoallv_packed_p2p(
  void const * send_buffer, int const * sendcounts, int const * sdispls,
  void * recv_buffer, int const * recvcounts, int const * rdispls,
  MPI_Comm comm) {

  yac_alltoallv_p2p(
    (void const *)send_buffer, sendcounts, sdispls,
    (void *)recv_buffer, recvcounts, rdispls, 1, MPI_PACKED, comm);
}

void yac_alltoallv_p2p_group(
  void const * send_buffer, int const * sendcounts, int const * sdispls,
  void * recv_buffer, int const * recvcounts, int const * rdispls,
  size_t dt_size, MPI_Datatype dt, struct yac_group_comm group_comm) {

  MPI_Comm comm = group_comm.comm;
  int comm_rank;
  yac_mpi_call(MPI_Comm_rank(comm, &comm_rank), comm);
  int rank = comm_rank - group_comm.start;

  int req_count = 0;
  for (int i = 0; i < group_comm.size; ++i)
    req_count += (sendcounts[i] > 0) + (recvcounts[i] > 0);
  MPI_Request * req = xmalloc((size_t)req_count * sizeof(*req));

  req_count = 0;
  for (int j = 0, lb = rank, ub = group_comm.size; j < 2;
       ++j, lb = 0, ub = rank) {
    for (int i = lb; i < ub; ++i) {
      if (sendcounts[i] > 0) {

        yac_mpi_call(
          MPI_Isend(
            (void const *)((unsigned char *)send_buffer +
                           dt_size * (size_t)(sdispls[i])),
            sendcounts[i], dt, i + group_comm.start, 0,
            comm, req + req_count), comm);
        ++req_count;
      }
      if (recvcounts[i] > 0) {
        yac_mpi_call(
          MPI_Irecv(
            (void *)((unsigned char *)recv_buffer +
                     dt_size * (size_t)(rdispls[i])),
            recvcounts[i], dt, i + group_comm.start, 0,
            comm, req + req_count), comm);
        ++req_count;
      }
    }
  }
  yac_mpi_call(MPI_Waitall(req_count, req, MPI_STATUSES_IGNORE), comm);
  free(req);
}

static int nearest_power_of_two(int x) {
  int power = 1;

  while(power < x) power *= 2;

  return power / 2;
}

// based on https://doi.org/10.1016/j.parco.2017.08.004
void yac_allreduce_sum_dble(
  double * buffer, int count, struct yac_group_comm group_comm) {

  int comm_rank;
  yac_mpi_call(MPI_Comm_rank(group_comm.comm, &comm_rank), group_comm.comm);

  int rank = comm_rank - group_comm.start;
  int pof2 = nearest_power_of_two(group_comm.size);
  int rem = group_comm.size - pof2;
  int my_rank;
  double * recv_buffer = xmalloc((size_t)count * sizeof(*recv_buffer));

  if (rank < 2 * rem) {

    if (rank & 1) {
      yac_mpi_call(
        MPI_Recv(
          (void*)recv_buffer, count, MPI_DOUBLE, rank - 1 + group_comm.start, 0,
          group_comm.comm, MPI_STATUS_IGNORE), group_comm.comm);
      for (int i = 0; i < count; ++i) buffer[i] += recv_buffer[i];
      my_rank = rank / 2;
    } else {
      yac_mpi_call(
        MPI_Send(
          (void const *)buffer, count, MPI_DOUBLE, rank + 1 + group_comm.start,
          0, group_comm.comm), group_comm.comm);
      my_rank = -1;
    }
  } else {
    my_rank = rank - rem;
  }
  if (my_rank != -1) {
    int mask = 1;
    while (mask < pof2) {
      int newdst = my_rank ^ mask;
      int dst;
      if (newdst < rem) dst = newdst * 2 + 1;
      else              dst = newdst + rem;
      yac_mpi_call(
        MPI_Sendrecv(
          (void const*)buffer, count, MPI_DOUBLE, dst + group_comm.start, 0,
          (void*)recv_buffer, count, MPI_DOUBLE, dst + group_comm.start, 0,
          group_comm.comm, MPI_STATUS_IGNORE),
        group_comm.comm);
      for (int i = 0; i < count; ++i) buffer[i] += recv_buffer[i];
      mask <<= 1;
    }
  }
  free(recv_buffer);
  if (rank < 2 * rem) {
    if (rank & 1) {
      yac_mpi_call(
        MPI_Send(
          (void const*)buffer, count, MPI_DOUBLE, rank - 1 + group_comm.start,
          0, group_comm.comm), group_comm.comm);
    } else {
      yac_mpi_call(
        MPI_Recv(
          (void*)buffer, count, MPI_DOUBLE, rank + 1 + group_comm.start, 0,
          group_comm.comm, MPI_STATUS_IGNORE), group_comm.comm);
    }
  }
}

static int log2_(int x) {
  if (x <= 1) return 0;
  int l2 = 0;
  while (x >>= 1) ++l2;
  return l2;
}

// based on https://doi.org/10.1109/71.642949
void yac_allgather_uint64(
  const uint64_t * sendbuf, uint64_t * recvbuf, int count,
  struct yac_group_comm group_comm) {

  int comm_rank;
  yac_mpi_call(MPI_Comm_rank(group_comm.comm, &comm_rank), group_comm.comm);
  int rank = comm_rank - group_comm.start;

  uint64_t * temp = xmalloc((size_t)group_comm.size * (size_t)count * sizeof(*temp));

  int lg2 = log2_(group_comm.size);
  memcpy(temp, sendbuf, (size_t)count * sizeof(*temp));
  int nblk = 1;
  int curr_len = count;

  for (int r = 0; r < lg2; ++r) {
    int dst = (rank - nblk + group_comm.size) % group_comm.size;
    int src = (rank + nblk) % group_comm.size;
    yac_mpi_call(
      MPI_Sendrecv(
        (void const*)temp, curr_len, MPI_UINT64_T, dst + group_comm.start, 0,
        (void *)(temp + (size_t)curr_len), curr_len, MPI_UINT64_T,
        src + group_comm.start, 0, group_comm.comm, MPI_STATUS_IGNORE),
      group_comm.comm);
    nblk *= 2;
    curr_len *= 2;
  }
  int rest = count * group_comm.size - curr_len;
  int dst = (rank - nblk + group_comm.size) % group_comm.size;
  int src = (rank + nblk) % group_comm.size;
  yac_mpi_call(
    MPI_Sendrecv(
      (void const*)temp, rest, MPI_UINT64_T, dst + group_comm.start, 0,
      (void*)(temp + (size_t)curr_len), rest, MPI_UINT64_T,
      src + group_comm.start, 0, group_comm.comm, MPI_STATUS_IGNORE),
    group_comm.comm);
  memcpy(recvbuf + (size_t)count * (size_t)rank,
         temp, (size_t)count * (size_t)(group_comm.size - rank) * sizeof(*temp));
  memcpy(recvbuf, temp + (size_t)count * (size_t)(group_comm.size - rank),
         (size_t)count * (size_t)rank * sizeof(*temp));

  free(temp);
}

void yac_bcast_group(
  void * buffer, int count, MPI_Datatype datatype, int root,
  struct yac_group_comm group_comm) {

  int comm_rank;
  yac_mpi_call(MPI_Comm_rank(group_comm.comm, &comm_rank), group_comm.comm);
  int rank = comm_rank - group_comm.start;

  // if root is not part of the group
  if ((root < group_comm.start) ||
      (root >= group_comm.start + group_comm.size)) {

    if (comm_rank == root) {
      yac_mpi_call(
        MPI_Send(
          (void const*)buffer, count, datatype, group_comm.start, 0,
          group_comm.comm), group_comm.comm);
      return;
    } else if (comm_rank == group_comm.start) {
      yac_mpi_call(
        MPI_Recv(
          buffer, count, datatype, root, 0, group_comm.comm,
          MPI_STATUS_IGNORE), group_comm.comm);
    }
    root = 0;
  } else {
    root -= group_comm.start;
  }

  // if not root, receive data
  if (rank != root) {

    int temp_rank = (group_comm.size + rank - root) % group_comm.size;
    int bit = 1;

    while (bit <= temp_rank) bit <<= 1;
    bit >>= 1;

    int src_rank =
      (((temp_rank ^ bit) + root) % group_comm.size) + group_comm.start;

    yac_mpi_call(
      MPI_Recv(buffer, count, datatype, src_rank, 0, group_comm.comm,
               MPI_STATUS_IGNORE), group_comm.comm);
  }

  // relative rank in respect to root
  int temp_rank = (group_comm.size + rank - root) % group_comm.size;
  int bit = 1, send_rank;

  while(bit <= temp_rank) bit <<= 1;

  while ((send_rank = temp_rank | bit) < group_comm.size) {

    bit <<= 1;

    send_rank = ((send_rank + root) % group_comm.size) + group_comm.start;

    yac_mpi_call(
      MPI_Send(
        (void const*)buffer, count, datatype, send_rank, 0, group_comm.comm),
      group_comm.comm);
  }
}

struct yac_group_comm yac_group_comm_new(MPI_Comm comm) {

  struct yac_group_comm group_comm;
  group_comm.start = 0;
  yac_mpi_call(MPI_Comm_size(comm, &(group_comm.size)), comm);
  yac_mpi_call(MPI_Comm_dup(comm, &(group_comm.comm)), comm);

  return group_comm;
}

void yac_group_comm_delete(struct yac_group_comm group_comm) {

  yac_mpi_call(MPI_Comm_free(&(group_comm.comm)), group_comm.comm);
}

int yac_group_comm_get_rank(struct yac_group_comm group_comm) {
  return yac_group_comm_get_global_rank(group_comm) - group_comm.start;
}

int yac_group_comm_get_size(struct yac_group_comm group_comm) {
  return group_comm.size;
}

int yac_group_comm_get_global_rank(struct yac_group_comm group_comm) {
  int comm_rank;
  yac_mpi_call(MPI_Comm_rank(group_comm.comm, &comm_rank), group_comm.comm);
  return comm_rank;
}

int yac_group_comm_get_global_size(struct yac_group_comm group_comm) {
  int comm_size;
  yac_mpi_call(MPI_Comm_size(group_comm.comm, &comm_size), group_comm.comm);
  return comm_size;
}

void yac_group_comm_split(
  struct yac_group_comm group_comm, int split_rank,
  struct yac_group_comm * local_group_comm,
  struct yac_group_comm * remote_group_comm) {

  int comm_rank;
  MPI_Comm comm = group_comm.comm;
  yac_mpi_call(MPI_Comm_rank(comm, &comm_rank), comm);

  YAC_ASSERT(
    (split_rank >= 0) && (split_rank < group_comm.size),
    "ERROR(yac_group_comm_split): invalid split rank")

  int start[2] = {group_comm.start, group_comm.start + split_rank};
  int size[2] = {split_rank, group_comm.size - split_rank};
  int local_idx = (comm_rank - group_comm.start) >= split_rank;

  local_group_comm->start = start[local_idx];
  local_group_comm->size = size[local_idx];
  local_group_comm->comm = comm;
  remote_group_comm->start = start[local_idx^1];
  remote_group_comm->size = size[local_idx^1];
  remote_group_comm->comm = comm;
}

MPI_Datatype yac_get_bounding_circle_mpi_datatype(MPI_Comm comm) {

  struct bounding_circle dummy;
  MPI_Datatype bnd_circle_dt;
  int array_of_blocklengths[] = {3, 1, 1};
  const MPI_Aint array_of_displacements[] =
    {(MPI_Aint)((void*)&(dummy.base_vector[0]) - (void*)&dummy),
     (MPI_Aint)((void*)&(dummy.inc_angle.sin) - (void*)&dummy),
     (MPI_Aint)((void*)&(dummy.inc_angle.cos) - (void*)&dummy)};
  const MPI_Datatype array_of_types[] =
    {MPI_DOUBLE, MPI_DOUBLE, MPI_DOUBLE};
  yac_mpi_call(
    MPI_Type_create_struct(3, array_of_blocklengths, array_of_displacements,
                           array_of_types, &bnd_circle_dt), comm);
  return yac_create_resized(bnd_circle_dt, sizeof(dummy), comm);
}

MPI_Datatype yac_create_resized(
  MPI_Datatype dt, size_t new_size, MPI_Comm comm) {

  MPI_Datatype resized_dt;

#define OPENMPI_WORKAROUND
#ifdef OPENMPI_WORKAROUND
  MPI_Aint lb, extent;
  MPI_Type_get_extent(dt, &lb, &extent);
  yac_mpi_call(
    MPI_Type_create_resized(dt, lb, (MPI_Aint)new_size, &resized_dt), comm);
#else
  yac_mpi_call(
    MPI_Type_create_resized(dt, 0, (MPI_Aint)new_size, &resized_dt), comm);
#endif
#undef OPENMPI_WORKAROUND
  yac_mpi_call(MPI_Type_free(&dt), comm);
  yac_mpi_call(MPI_Type_commit(&resized_dt), comm);
  return resized_dt;
}

void yac_generate_alltoallv_args(
  int count, int const * sendcounts, int * recvcounts,
  int * sdispls, int * rdispls, MPI_Comm comm) {

  int comm_size;
  yac_mpi_call(MPI_Comm_size(comm, &comm_size), comm);

  // exchange the number of requested points
  yac_mpi_call(
    MPI_Alltoall(
      (void const*)sendcounts, count, MPI_INT,
      (void*)recvcounts, count, MPI_INT, comm), comm);

  // sdispls are offset by one position, this is intentional because this is
  // usefull for packing of data
  sdispls[0] = 0;
  for (int i = 0, saccu = 0, raccu = 0; i < count * comm_size; ++i) {
    sdispls[i+1] = saccu;
    rdispls[i] = raccu;
    saccu += sendcounts[i];
    raccu += recvcounts[i];
  }
}

void yac_get_comm_buffers(
  int count, int ** sendcounts, int ** recvcounts,
  int ** sdispls, int ** rdispls, MPI_Comm comm) {

  int comm_size;
  yac_mpi_call(MPI_Comm_size(comm, &comm_size), comm);

  int * comm_buffer_;
  if (!comm_buffer_in_use) {
    ENSURE_ARRAY_SIZE(
      comm_buffer, comm_buffer_array_size,
      4 * (size_t)count * (size_t)comm_size + 1);
    comm_buffer_ = comm_buffer;
    comm_buffer_in_use = 1;
  } else {
    comm_buffer_ =
      xmalloc(
        (4 * (size_t)count * (size_t)comm_size + 1) * sizeof(*comm_buffer_));
  }

  size_t offset = (size_t)count * (size_t)comm_size;
  *sendcounts = comm_buffer_ + 0 * offset;
  *recvcounts = comm_buffer_ + 1 * offset;
  *rdispls =    comm_buffer_ + 2 * offset;
  *sdispls =    comm_buffer_ + 3 * offset; // sdispls is bigger by one element,
                                       // which is usefull when packing data
                                       // for alltoallv operation
  memset(
    comm_buffer_, 0, (size_t)count * (size_t)comm_size * sizeof(*comm_buffer_));
}

void yac_free_comm_buffers(
  int * sendcounts, int * recvcounts, int * sdispls, int * rdispls) {

  if (sendcounts != comm_buffer) free(sendcounts);
  else comm_buffer_in_use = 0;
}

void yac_mpi_init(MPI_Comm comm) {

  if (init_count == 0) {
    int mpi_initialized;
    MPI_Initialized(&mpi_initialized);

    if (!mpi_initialized) {
      MPI_Init(NULL, NULL);
      mpi_initialised_by_yac = 1;
    }

    if (!xt_initialized() || xt_finalized()) {
      xt_initialize(comm);
      yaxt_initialised_by_yac = 1;
    }
  } else {

    YAC_ASSERT(
      !yaxt_initialised_by_yac,
      "ERROR(yac_mpi_init): YAXT was initialised by YAC. \n"
      "In case there are multiple instances of YAC in parallel, the user has "
      "to initialise YAXT such that it is available on all processes that "
      "use YAC.")
  }
  init_count++;
}

void yac_mpi_cleanup() {

  if (init_count == 1) {
    YAC_ASSERT(
      !comm_buffer_in_use, "ERROR(yac_mpi_finalize): comm_buffer still in use")
    free(comm_buffer);
    comm_buffer = NULL;
    comm_buffer_array_size = 0;
    if (yaxt_initialised_by_yac) {
      xt_finalize();
      yaxt_initialised_by_yac = 0;
    }
  }
  init_count--;
}

void yac_mpi_finalize() {

  yac_mpi_cleanup();
  if (mpi_initialised_by_yac)
    yac_mpi_call(MPI_Finalize(), MPI_COMM_WORLD);
}

/*
 * Local Variables:
 * c-basic-offset: 2
 * coding: utf-8
 * indent-tabs-mode: nil
 * show-trailing-whitespace: t
 * require-trailing-newline: t
 * End:
 */
