!
! @file test_dummy_coupling6_real.f90
!
! @copyright Copyright  (C)  2022 DKRZ, MPI-M
!
! @author Moritz Hanke <hanke@dkrz.de>
!         Rene Redler  <rene.redler@mpimet.mpg.de>
!
!
!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
! URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
!
! This file is part of YAC.
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!

#include "test_macros.inc"

PROGRAM main

  USE utest
  USE mo_yac_finterface
  USE yaxt
  USE mpi

  IMPLICIT NONE

  INTEGER, PARAMETER :: sp = SELECTED_REAL_KIND(6, 37) !< single precision
  INTEGER, PARAMETER :: wp = sp                        !< selected working precision

  ! M_PI / 180
  REAL(wp), PARAMETER :: YAC_RAD = 0.017453292519943295769_wp
  INTEGER :: rank, ierror

  ! ===================================================================

  CALL MPI_Init(ierror)
  CALL xt_initialize(MPI_COMM_WORLD)
  CALL start_test('dummy_coupling6')

  CALL MPI_Comm_rank(MPI_COMM_WORLD, rank, ierror)

  SELECT CASE (rank)
    CASE (0)
      CALL yac_finit_comm_dummy(MPI_COMM_WORLD)
      CALL yac_ffinalize()
    CASE (1)
      CALL run_comp_a()
    CASE (2)
      CALL run_comp_b()
    CASE DEFAULT
      CALL test(.FALSE.)
  END SELECT

  CALL xt_finalize()
  CALL MPI_Finalize(ierror)

  CALL stop_test
  CALL exit_tests

CONTAINS

  SUBROUTINE check_pair_comm(pair_comm, proc_name)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: pair_comm

    CHARACTER, INTENT(IN) :: proc_name

    CHARACTER(LEN=1) :: sendbuf
    CHARACTER(LEN=2) :: recvbuf
    INTEGER :: ierror

    sendbuf(1:1) = proc_name

    CALL MPI_Allgather( &
      sendbuf, 1, MPI_CHARACTER, recvbuf, 1, MPI_CHARACTER, pair_comm, ierror)

    CALL test(recvbuf == "AB")

  END SUBROUTINE

  SUBROUTINE run_comp_a()

    IMPLICIT NONE

    INTEGER :: comp_id, grid_id, point_id, field_id
    INTEGER :: pair_comm

    INTEGER, PARAMETER :: nbr_vertices = 4
    INTEGER, PARAMETER :: nbr_cells = 2
    INTEGER, DIMENSION(2), PARAMETER :: nbr_vertices_per_cell = (/3,3/)
    REAL(wp), DIMENSION(4), PARAMETER :: &
      x_vertices =                               &
        (/0.0 * YAC_RAD, -1.0 * YAC_RAD, 1.0 * YAC_RAD, 0.0 * YAC_RAD/)
    REAL(wp), DIMENSION(4), PARAMETER :: &
      y_vertices =                               &
        (/1.0 * YAC_RAD, 0.0 * YAC_RAD, 0.0 * YAC_RAD, -1.0 * YAC_RAD/)
    INTEGER, DIMENSION(6), PARAMETER :: cell_to_vertex = (/1,2,3, 2,4,3/)
    REAL(wp), DIMENSION(2), PARAMETER :: &
      x_cells = (/0.0 * YAC_RAD, 0.0 * YAC_RAD/)
    REAL(wp), DIMENSION(2), PARAMETER :: &
      y_cells = (/-0.5 * YAC_RAD, 0.5 * YAC_RAD/)

    INTEGER :: nbr_fields
    INTEGER, DIMENSION(1) :: field_ids
    INTEGER, DIMENSION(1) :: roles

    INTEGER, PARAMETER :: nbr_hor_points = 2
    INTEGER, PARAMETER :: collection_size = 1
    INTEGER, PARAMETER :: nbr_pointsets = 1
    INTEGER :: info

    ! initialise YAC default instance
    CALL yac_finit('coupling_test6.xml', 'coupling.xsd')

    ! define local component
    CALL yac_fdef_comp('comp_A', comp_id)

    ! get communicator that contains both components
    CALL yac_fget_pair_rootcomm(pair_comm, 'comp_A', 'comp_B')

    ! check the pair_comm
    CALL check_pair_comm(pair_comm, 'A')
    CALL MPI_Comm_free(pair_comm, ierror)

    !   define local grid
    !
    !      0
    !     / \
    !    / o \
    !   /     \
    !  1-------2   Eq.
    !   \     /
    !    \ o /
    !     \ /
    !      3
    CALL yac_fdef_grid(                                              &
      'grid_A', nbr_vertices, nbr_cells, SUM(nbr_vertices_per_cell), &
      nbr_vertices_per_cell, x_vertices, y_vertices, cell_to_vertex, grid_id)

    ! center points in cells (needed e.g. for nearest neighbour)
    CALL yac_fdef_points( &
      grid_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells, point_id)

    ! define field
    CALL yac_fdef_field('A_to_B', comp_id, (/point_id/), 1, field_id)

    ! generate coupling
    CALL yac_fsearch(ierror)

    !-----------------------------------------------------
    ! some check
    !-----------------------------------------------------
    CALL yac_fget_nbr_fields(nbr_fields)
    CALL yac_fget_field_ids(1, field_ids)
    CALL test(nbr_fields == 1)
    CALL test(ALL(field_ids(1:1) == (/field_id/)))

    CALL yac_fget_field_id_from_name('A_to_B', field_ids(1))
    CALL test(ALL(field_ids(1:1) == (/field_id/)))

    CALL yac_fget_field_id_from_field_and_grid_name( &
      'A_to_B', 'grid_A', field_ids(1))
    CALL test(ALL(field_ids(1:1) == (/field_id/)))

    CALL yac_fget_role_of_field('A_to_B', roles(1))
    CALL test(ALL(roles(1:1) == (/1/)))

    CALL yac_fget_role_of_field( &
      'A_to_B', 'grid_A', roles(1))
    CALL test(ALL(roles(1:1) == (/1/)))
    !-----------------------------------------------------

    ! move data from comp_A to comp_B
    CALL yac_fput(                                                               &
      field_id, nbr_hor_points, collection_size, nbr_pointsets,                  &
      RESHAPE((/3.0, 4.0/), (/nbr_hor_points, nbr_pointsets, collection_size/)), &
      info, ierror)

    CALL yac_ffinalize()

  END SUBROUTINE run_comp_a

  SUBROUTINE run_comp_b()

    IMPLICIT NONE

    INTEGER :: default_instance_id, comp_id, grid_id_B, point_id_B, field_id
    INTEGER :: pair_comm

    INTEGER :: internal_comm
    INTEGER :: internal_instance_id, internal_comp_id, grid_id_C, point_id_C
    INTEGER :: field_id_src, field_id_tgt

    INTEGER, DIMENSION(2), PARAMETER :: nbr_vertices = (/2,3/)
    INTEGER, DIMENSION(2), PARAMETER :: cyclic = (/0,0/)
    INTEGER, DIMENSION(2), PARAMETER :: nbr_cells = (/1,2/)
    REAL(wp), DIMENSION(2), PARAMETER :: &
      x_vertices = (/-0.5 * YAC_RAD, 0.5 * YAC_RAD/)
    REAL(wp), DIMENSION(3), PARAMETER :: &
      y_vertices = (/-1.0 * YAC_RAD, 0.0 * YAC_RAD, 1.0 * YAC_RAD/)
    REAL(wp), DIMENSION(1), PARAMETER :: &
      x_cells = (/0.0 * YAC_RAD/)
    REAL(wp), DIMENSION(2), PARAMETER :: &
      y_cells = (/-0.5 * YAC_RAD, 0.5 * YAC_RAD/)

    REAL(wp), DIMENSION(2), PARAMETER :: &
      internal_x_vertices = (/-0.4 * YAC_RAD, 0.4 * YAC_RAD/)
    REAL(wp), DIMENSION(3), PARAMETER :: &
      internal_y_vertices = (/-0.7 * YAC_RAD, 0.0 * YAC_RAD, 0.8 * YAC_RAD/)

    INTEGER :: nbr_fields
    INTEGER, DIMENSION(2) :: field_ids
    INTEGER, DIMENSION(2) :: roles

    INTEGER, PARAMETER :: nbr_hor_points = 2
    INTEGER, PARAMETER :: nbr_pointsets = 1
    INTEGER, PARAMETER :: collection_size = 1
    REAL(wp), DIMENSION(nbr_hor_points,collection_size) :: recv_field_B
    INTEGER :: info

    REAL(wp), DIMENSION(6,collection_size) :: recv_field_C
    INTEGER :: send_info, recv_info

    ! initialise YAC default instance
    CALL yac_finit('coupling_test6.xml', 'coupling.xsd', default_instance_id)

    ! define local component
    CALL yac_fdef_comp(default_instance_id, 'comp_B', comp_id)

    ! get communicator that contains both components
    CALL yac_fget_pair_rootcomm( &
      default_instance_id, pair_comm, 'comp_A', 'comp_B')

    ! check the pair_comm
    CALL check_pair_comm(pair_comm, 'B')
    CALL MPI_Comm_free(pair_comm, ierror)

    !   define local grid
    !
    !   4-------5
    !   |       |
    !   |   o   |
    !   |       |
    !   2-------3
    !   |       |
    !   |   o   |
    !   |       |
    !   0-------1
    CALL yac_fdef_grid( &
      'grid_B', nbr_vertices, cyclic, x_vertices, y_vertices, grid_id_B)

    ! center points in cells (needed e.g. for nearest neighbour)
    CALL yac_fdef_points( &
      grid_id_B, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells, point_id_B)

    ! define field
    CALL yac_fdef_field('A_to_B', comp_id, (/point_id_B/), 1, field_id)

    ! generate coupling
    CALL yac_fsearch(default_instance_id, ierror)

    !-----------------------------------------------------
    ! some check
    !-----------------------------------------------------
      CALL yac_fget_nbr_fields(default_instance_id, nbr_fields)
      CALL yac_fget_field_ids(default_instance_id, 1, field_ids)
      CALL test(nbr_fields == 1)
      CALL test(ALL(field_ids(1:1) == (/field_id/)))

      CALL yac_fget_field_id_from_name( &
        default_instance_id, 'A_to_B', field_ids(1))
      CALL test(ALL(field_ids(1:1) == (/field_id/)))

      CALL yac_fget_field_id_from_field_and_grid_name( &
        default_instance_id, 'A_to_B', 'grid_B', field_ids(1))
      CALL test(ALL(field_ids(1:1) == (/field_id/)))

      CALL yac_fget_role_of_field(default_instance_id, 'A_to_B', roles(1))
      CALL test(ALL(roles(1:1) == (/2/)))

      CALL yac_fget_role_of_field( &
        default_instance_id, 'A_to_B', 'grid_B', roles(1))
      CALL test(ALL(roles(1:1) == (/2/)))
    !-----------------------------------------------------

    ! move data from comp_A to comp_B
    CALL yac_fget( &
      field_id, nbr_hor_points, collection_size, recv_field_B, info, ierror)

    !-----------------------------------------------------
    ! run internal instance
    !-----------------------------------------------------

      CALL yac_fget_localcomm(internal_comm, comp_id)

      ! initialise internal YAC instance
      CALL yac_finit_comm(                          &
        'coupling_test6_local.xml', 'coupling.xsd', &
        internal_comm, internal_instance_id)

      ! define internal component
      CALL yac_fdef_comp(internal_instance_id, 'comp_B', internal_comp_id)

      !   define local grid
      !
      !   4-------5
      !   |       |
      !   |   o   |
      !   |       |
      !   2-------3
      !   |       |
      !   |   o   |
      !   |       |
      !   0-------1
      CALL yac_fdef_grid(               &
        'grid_C', nbr_vertices, cyclic, &
        internal_x_vertices, internal_y_vertices, grid_id_C)

      ! vertices points (needed e.g. for nearest neighbour)
      CALL yac_fdef_points(                           &
        grid_id_C, nbr_vertices, YAC_LOCATION_CORNER, &
        internal_x_vertices, internal_y_vertices, point_id_C)

      ! define field
      CALL yac_fdef_field( &
        'B_to_C', internal_comp_id, (/point_id_B/), 1, field_id_src)
      CALL yac_fdef_field( &
        'B_to_C', internal_comp_id, (/point_id_C/), 1, field_id_tgt)

      ! generate internal coupling
      CALL yac_fsearch(internal_instance_id, ierror)

      !-----------------------------------------------------
      ! some check
      !-----------------------------------------------------
        CALL yac_fget_nbr_fields(internal_instance_id, nbr_fields)
        CALL yac_fget_field_ids(internal_instance_id, 2, field_ids)
        CALL test(nbr_fields == 2)
        CALL test(ALL(field_ids(1:2) == (/field_id_src, field_id_tgt/)))

        CALL yac_fget_field_id_from_field_and_grid_name( &
          internal_instance_id, 'B_to_C', 'grid_B', field_ids(1))
        CALL yac_fget_field_id_from_field_and_grid_name( &
          internal_instance_id, 'B_to_C', 'grid_C', field_ids(2))
        CALL test(ALL(field_ids(1:2) == (/field_id_src, field_id_tgt/)))

        CALL yac_fget_role_of_field( &
          internal_instance_id, 'B_to_C', 'grid_B', roles(1))
        CALL yac_fget_role_of_field( &
          internal_instance_id, 'B_to_C', 'grid_C', roles(2))
        CALL test(ALL(roles(1:2) == (/1,2/)))
      !-----------------------------------------------------

      ! move data from grid B to grid C
      CALL yac_fexchange( &
        field_id_src, field_id_tgt, nbr_hor_points, nbr_pointsets, 6,        &
        collection_size,                                                  &
        RESHAPE(                                                             &
          recv_field_B, (/nbr_hor_points, nbr_pointsets, collection_size/)), &
        recv_field_C, send_info, recv_info, ierror)

      CALL test(ALL(RESHAPE(recv_field_C, (/6/)) == (/3.0, 3.0, 3.0, 3.0, 4.0, 4.0/)))

      CALL yac_fcleanup(internal_instance_id)

      CALL MPI_Comm_free(internal_comm, ierror)

    !-----------------------------------------------------

    CALL yac_ffinalize(default_instance_id)

  END SUBROUTINE

END PROGRAM
