/**
 * @file test_restart.c
 *
 * @copyright Copyright  (C)  2021 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#include <mpi.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "yac_interface.h"
#include "generate_cubed_sphere.h"
#include "test_function.h"
#include "utils.h"

#define DUMMY_VALUE (-1337.0)

static void run_setup(char local_comp_name, char comp_groups[2][2]);

int main (void) {

  // This test checks the restarting of YAC using varying MPI communicators and XML
  // configuration files.
  // This setup could also be executed with a single YAC initilisation.

  MPI_Init(NULL, NULL);

  int comm_rank, comm_size;
  MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);

  if (comm_size != 15) {
    fputs("wrong number of processes (has to be 15)\n", stderr);
    exit(EXIT_FAILURE);
  }

  char local_comp_name;
  if (comm_rank < 6)       local_comp_name = 'A';
  else if (comm_rank < 10) local_comp_name = 'B';
  else if (comm_rank < 13) local_comp_name = 'C';
  else                     local_comp_name = 'D';

  // couple A-B and C-D (each in their own world comm)
  run_setup(local_comp_name, (char[2][2]){{'A','B'},{'C','D'}});

  // couple A-C and B-D (each in their own world comm)
  run_setup(local_comp_name, (char[2][2]){{'A','C'},{'B','D'}});

  // couple A-D and B-C (each in their own world comm)
  run_setup(local_comp_name, (char[2][2]){{'A','D'},{'B','C'}});

  MPI_Finalize();

  exit(EXIT_SUCCESS);
}

static void def_grid(
  char const * grid_name, int comm_rank, int comm_size,
  int * grid_id, int * cell_point_id, int local_test_func_idx,
  int remote_test_func_idx, double ** field_data_out,
  double ** ref_field_data_in, size_t * field_data_size) {

  unsigned n = 50;

  unsigned nbr_vertices;
  unsigned nbr_cells;
  unsigned * num_vertices_per_cell;
  unsigned * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_core_mask;
  int * corner_core_mask;
  int * global_cell_id;
  int * global_corner_id;

  generate_part_cube_grid_information(
    n, &nbr_vertices, &nbr_cells, &num_vertices_per_cell, &cell_to_vertex,
    &x_vertices, &y_vertices, &x_cells, &y_cells, &global_cell_id,
    &cell_core_mask, &global_corner_id, &corner_core_mask,
    comm_rank, comm_size);

  yac_cdef_grid_unstruct(
    grid_name, nbr_vertices, nbr_cells, (int*)num_vertices_per_cell,
    x_vertices, y_vertices, (int*)cell_to_vertex, grid_id);

  yac_cset_global_index(global_cell_id, YAC_LOCATION_CELL, *grid_id);
  yac_cset_core_mask(cell_core_mask, YAC_LOCATION_CELL, *grid_id);

  yac_cdef_points_unstruct(
    *grid_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells,
    cell_point_id);

  double (*test_func[4])(double, double) =
    {&test_ana_fcos, &test_vortex, &test_gulfstream, &test_harmonic};

  *field_data_out = xmalloc(nbr_cells * sizeof(**field_data_out));
  for (int i = 0; i < nbr_cells; ++i)
    (*field_data_out)[i] =
      (cell_core_mask[i])?
        (*(test_func[local_test_func_idx]))(x_cells[i], y_cells[i]):
        DUMMY_VALUE;
  *ref_field_data_in = xmalloc(nbr_cells * sizeof(**ref_field_data_in));
  for (int i = 0; i < nbr_cells; ++i)
    (*ref_field_data_in)[i] =
      (cell_core_mask[i])?
        (*(test_func[remote_test_func_idx]))(x_cells[i], y_cells[i]):
        DUMMY_VALUE;
  *field_data_size = nbr_cells;
  free(cell_core_mask);
  free(corner_core_mask);
  free(global_cell_id);
  free(global_corner_id);
  free(x_vertices);
  free(y_vertices);
  free(x_cells);
  free(y_cells);
  free(num_vertices_per_cell);
  free(cell_to_vertex);
}

static void generate_xml_file(
  char comp_names[2], char const * xml_file_name) {

  char const * xml_config_file_format =
#include "test_restart2_xml_string.h"
;

  // generate xml file
  FILE * xml_config_file = xfopen(xml_file_name, "w+");
  fprintf(xml_config_file, xml_config_file_format,
          comp_names[0], comp_names[1],
          comp_names[0], comp_names[1],
          comp_names[1], comp_names[0],
          comp_names[0], comp_names[1]);
  xfclose(xml_config_file);
}

static void run_component(
  char local_comp_name, char remote_comp_name, MPI_Comm comm) {

  int comm_rank, comm_size;
  MPI_Comm_rank(comm, &comm_rank);
  MPI_Comm_size(comm, &comm_size);

  char comp_names[2];
  if (local_comp_name < remote_comp_name) {
    comp_names[0] = local_comp_name;
    comp_names[1] = remote_comp_name;
  } else {
    comp_names[0] = remote_comp_name;
    comp_names[1] = local_comp_name;
  }

  char xml_file_name[64];
  sprintf(
    xml_file_name, "test_restart2_%c_%c.xml", comp_names[0], comp_names[1]);

  // generate xml file
  if (comm_rank == 0) generate_xml_file(comp_names, xml_file_name);

  MPI_Barrier(comm);

  // init YAC
  yac_cinit_comm(xml_file_name, "coupling.xsd", comm);

  // define component
  char comp_name[16];
  int comp_id;
  sprintf(comp_name, "comp%c", local_comp_name);
  yac_cdef_comp(comp_name, &comp_id);

  MPI_Comm comp_comm;
  int comp_rank, comp_size;
  yac_cget_localcomm(&comp_comm, comp_id);
  MPI_Comm_rank(comp_comm, &comp_rank);
  MPI_Comm_size(comp_comm, &comp_size);
  MPI_Comm_free(&comp_comm);

  // define grid
  char grid_name[16];
  int grid_id, cell_point_id;
  double * field_data_out, * ref_field_data_in;
  size_t field_data_size;
  sprintf(grid_name, "grid%c", local_comp_name);
  def_grid(grid_name, comp_rank, comp_size, &grid_id, &cell_point_id,
           (int)(local_comp_name - 'A'), (int)(remote_comp_name - 'A'),
           &field_data_out, &ref_field_data_in, &field_data_size);

  // define fields
  char field_out_name[16];
  char field_in_name[16];
  int field_out_id, field_in_id;
  sprintf(field_out_name, "field_%c_to_%c", local_comp_name, remote_comp_name);
  sprintf(field_in_name, "field_%c_to_%c", remote_comp_name, local_comp_name);
  yac_cdef_field(
    field_out_name, comp_id, &cell_point_id, 1, &field_out_id);
  yac_cdef_field(
    field_in_name, comp_id, &cell_point_id, 1, &field_in_id);

  int ierror;
  yac_csearch(&ierror);

  double * field_data_in = xmalloc(field_data_size * sizeof(*field_data_in));

  // do some ping-pongs
  for (int t = 0; t < 10; ++t) {

    {
      int info, err;
      double *point_set_data[1];
      double **collection_data[1] = {point_set_data};
      point_set_data[0] = field_data_out;
      yac_cput(field_out_id, 1, collection_data, &info, &err);
    }

    {
      for (size_t i = 0; i < field_data_size; ++i)
        field_data_in[i] = DUMMY_VALUE;

      int info, err;
      double *collection_data[1] = {field_data_in};
      yac_cget(field_in_id, 1, collection_data, &info, &err);

      for (size_t i = 0; i < field_data_size; ++i) {
        if(fabs(field_data_in[i] - ref_field_data_in[i]) > 1e-3) {
          fputs("data data_mismatch\n", stderr);
          exit(EXIT_FAILURE);
        }
      }
    }
  }
  free(field_data_in);
  free(field_data_out);
  free(ref_field_data_in);

  yac_cfinalize();

  MPI_Barrier(comm);
  if (comm_rank == 0) unlink(xml_file_name);
}


static void run_setup(char local_comp_name, char comp_groups[2][2]) {

  int group_idx = (local_comp_name == comp_groups[1][0]) ||
                  (local_comp_name == comp_groups[1][1]);

  MPI_Comm group_comm;
  MPI_Comm_split(MPI_COMM_WORLD, group_idx, 0, &group_comm);

  char remote_comp_name =
    comp_groups[group_idx][comp_groups[group_idx][0] == local_comp_name];

  run_component(local_comp_name, remote_comp_name, group_comm);

  MPI_Comm_free(&group_comm);
}
